# SPDX-License-Identifier: AGPL-3.0-or-later
"""Framework for installing and updating distribution packages."""

import enum
import logging
import pathlib
import time

import apt
import apt.cache
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy, gettext_noop

import plinth.privileged.packages as privileged
from plinth import app as app_module
from plinth.diagnostic_check import (DiagnosticCheck,
                                     DiagnosticCheckParameters, Result)
from plinth.errors import MissingPackageError
from plinth.utils import format_lazy

from . import kvstore
from . import operation as operation_module
from .errors import PackageNotInstalledError

logger = logging.getLogger(__name__)


class PackageExpression:

    def possible(self) -> list[str]:
        """Return the list of possible packages before resolving."""
        raise NotImplementedError

    def actual(self) -> str:
        """Return the name of the package to install.

        TODO: Also return version and suite to install from.
        """
        raise NotImplementedError


class Package(PackageExpression):

    def __init__(
            self,
            name,
            optional: bool = False,
            version: str | None = None,  # ">=1.0,<2.0"
            distribution: str | None = None,  # Debian, Ubuntu
            suite: str | None = None,  # stable, testing
            codename: str | None = None,  # bullseye-backports
            architecture: str | None = None):  # arm64
        self.name = name
        self.optional = optional
        self.version = version
        self.distribution = distribution
        self.suite = suite
        self.codename = codename
        self.architecture = architecture

    def __repr__(self):
        return self.name

    def __or__(self, other):
        return PackageOr(self, other)

    def possible(self) -> list[str]:
        return [self.name]

    def actual(self) -> str:
        cache = apt.Cache()
        if self.name in cache and cache[self.name].candidate:
            # cache[package].candidate returns installation candidate. If the
            # package is not installable due to the available versions being of
            # priority less than 0, then .candidate will be None. TODO: Also
            # return version and suite to install from.
            return self.name

        raise MissingPackageError(self.name)


class PackageOr(PackageExpression):
    """Specify that one of the two packages will be installed."""

    def __init__(self, package1: PackageExpression,
                 package2: PackageExpression):
        self.package1 = package1
        self.package2 = package2

    def __repr__(self):
        return self.package1.name + ' | ' + self.package2.name

    def possible(self) -> list[str]:
        return self.package1.possible() + self.package2.possible()

    def actual(self) -> str:
        try:
            return self.package1.actual()
        except MissingPackageError:
            return self.package2.actual()


class Packages(app_module.FollowerComponent):
    """Component to manage the packages of an app.

    This component is responsible for installation, upgrades and uninstallation
    of packages required by an app.
    """

    class ConflictsAction(enum.Enum):
        """Action to take when a conflicting package is installed."""

        IGNORE = 'ignore'  # Proceed as if there are no conflicts
        REMOVE = 'remove'  # Remove the packages before installing the app

    def __init__(self, component_id: str,
                 packages: list[str | PackageExpression],
                 skip_recommends: bool = False,
                 conflicts: list[str] | None = None,
                 conflicts_action: ConflictsAction | None = None,
                 rerun_setup_on_upgrade: bool = False):
        """Initialize a new packages component.

        'component_id' should be a unique ID across all components of an app
        and across all components.

        'packages' is the list of Debian packages managed by this component.

        'skip_recommends' is a boolean specifying whether recommended packages
        should be installed along with the listed packages.

        'conflicts' is the list of Debian packages that can't simultaneously be
        installed with packages listed here. None if there are no known
        conflicting packages.

        'conflicts_action' is a string representing the action to take when it
        is found that conflicting Debian packages are installed on the system.
        None if there are no known conflicting packages.

        'rerun_setup_on_upgrade' is a boolean to specify whether app's setup
        process should be rerun if a .deb package is upgraded (by someone other
        than FreedomBox process).
        """
        super().__init__(component_id)

        self.component_id = component_id
        self._packages: list[PackageExpression] = []
        for package in packages:
            if isinstance(package, str):
                self._packages.append(Package(package))
            else:
                self._packages.append(package)

        self.skip_recommends = skip_recommends
        self.conflicts = conflicts or []
        self.conflicts_action = conflicts_action
        self.rerun_setup_on_upgrade = rerun_setup_on_upgrade

    @property
    def package_expressions(self) -> list[PackageExpression]:
        """Return the list of managed packages as expressions."""
        return self._packages

    @property
    def possible_packages(self) -> list[str]:
        """Return the list of possible packages before resolving."""
        packages: list[str] = []
        for package_expression in self.package_expressions:
            packages.extend(package_expression.possible())

        return packages

    def get_actual_packages(self) -> list[str]:
        """Return the computed list of packages to install.

        Raise MissingPackageError if a required package is not available.
        """
        return [
            package_expression.actual()
            for package_expression in self.package_expressions
        ]

    def setup(self, old_version):
        """Install the packages."""
        packages_to_remove = self.find_conflicts()
        if packages_to_remove and \
           self.conflicts_action not in (None, self.ConflictsAction.IGNORE):
            logger.info('Removing conflicting packages: %s',
                        packages_to_remove)
            uninstall(packages_to_remove, purge=False)

        install(self.get_actual_packages(),
                skip_recommends=self.skip_recommends)

    def uninstall(self):
        """Uninstall and purge the packages."""
        # Ensure package list is update-to-date before looking at dependencies.
        refresh_package_lists()

        # List of packages to purge from the system. Be resilient to a package
        # not being found in apt's cache by not using get_actual_packages(). If
        # a package expression is listed as (package1 | package2) then try to
        # remove both packages.
        packages = self.possible_packages
        logger.info('App\'s list of packages to remove: %s', packages)

        packages = self._filter_packages_to_keep(packages)
        uninstall(packages, purge=True)

    def diagnose(self) -> list[DiagnosticCheck]:
        """Run diagnostics and return results."""
        results = super().diagnose()
        cache = apt.Cache()
        for package_expression in self.package_expressions:
            try:
                package_name = package_expression.actual()
            except MissingPackageError:
                check_id = f'package-available-{package_expression}'
                description = gettext_noop('Package {package_expression} is '
                                           'not available for install')
                parameters: DiagnosticCheckParameters = {
                    'package_expression': str(package_expression)
                }
                results.append(
                    DiagnosticCheck(check_id, description, Result.FAILED,
                                    parameters, self.component_id))
                continue

            result = Result.WARNING
            latest_version = '?'
            if package_name in cache:
                package = cache[package_name]
                if package.candidate:
                    latest_version = package.candidate.version
                    if package.candidate.is_installed:
                        result = Result.PASSED

            check_id = f'package-latest-{package_name}'
            description = gettext_noop('Package {package_name} is the latest '
                                       'version ({latest_version})')
            parameters = {
                'package_name': str(package_name),
                'latest_version': str(latest_version)
            }
            results.append(
                DiagnosticCheck(check_id, description, result, parameters,
                                self.component_id))

        return results

    def find_conflicts(self) -> list[str] | None:
        """Return list of conflicting packages installed on the system."""
        if not self.conflicts:
            return None

        return packages_installed(self.conflicts)

    def is_available(self) -> bool:
        """Return whether all of the packages are available.

        Returns True if all of the packages are available in the user's Debian
        distribution or False otherwise. Returns True if it cannot be reliably
        determined whether the packages are available or not.

        This operation can run for a long time as package cache may be
        refreshed.
        """
        try:
            self.get_actual_packages()
            return True  # All packages were found in the cache
        except MissingPackageError:
            pass  # We will retry after refreshing package list

        package_cache = pathlib.Path('/var/cache/apt/pkgcache.bin')
        if package_cache.exists():
            package_cache_mtime = package_cache.stat().st_mtime

            sources_dir = pathlib.Path('/etc/apt/sources.list.d')
            source_files = ([pathlib.Path('/etc/apt/sources.list')] +
                            list(sources_dir.glob('*.list')) +
                            list(sources_dir.glob('*.sources')))
            are_sources_newer = any(
                source_file.stat().st_mtime > package_cache_mtime
                for source_file in source_files)

            # If the package cache is new, then package is really not
            # available.
            if (time.time() - package_cache.stat().st_mtime < 3600
                    and not are_sources_newer):
                return False

        # Perform 'apt-get update'
        refresh_package_lists()

        # Check again for all the packages
        try:
            self.get_actual_packages()
        except MissingPackageError:
            return False

        return True

    def _filter_packages_to_keep(self, packages: list[str]) -> list[str]:
        """Filter out the list of packages to keep from given list.

        Packages to keep are packages needed by other installed apps and their
        dependencies (PreDepends, Depends, Recommends).
        """
        packages_set: set[str] = set(packages)

        # Get list of packages needed by other installed apps and by freedombox
        # itself (packages to keep).
        keep_packages: set[str] = {'freedombox'}
        for app in app_module.App.list():
            # uninstall() will be called on Packages of this app separately
            # for uninstalling this app.
            if app == self.app:
                continue

            if app.get_setup_state() == app_module.App.SetupState.NEEDS_SETUP:
                continue

            # Remove packages used by other installed apps
            for component in app.get_components_of_type(Packages):
                keep_packages |= set(component.possible_packages)

        # Get list of all the dependencies of packages to keep.
        keep_packages_with_deps: set[str] = set()
        cache = apt.Cache()
        while keep_packages:
            package_name = keep_packages.pop()
            if package_name in keep_packages_with_deps:
                continue  # Already processed

            keep_packages_with_deps.add(package_name)
            if package_name not in cache:
                continue  # Package is not available in sources

            if not cache[package_name].is_installed:
                continue  # Package is not installed

            version = cache[package_name].installed
            if not version:
                continue

            dependencies = version.dependencies + version.recommends
            for dependency in dependencies:
                for or_dependency in dependency.or_dependencies:
                    keep_packages.add(or_dependency.name)

        # Filter out any packages that are to be kept or their dependencies.
        packages_set -= keep_packages_with_deps

        # Preserve order of packages for ease of testing.
        return [package for package in packages if package in packages_set]


class PackageException(Exception):
    """A package operation has failed."""


class Transaction:
    """Information about an ongoing transaction."""

    def __init__(self, app_id, package_names):
        """Initialize transaction object.

        Set most values to None until they are sent as progress update.
        """
        self.app_id = app_id
        self.package_names = package_names

        self._reset_status()

    def get_id(self):
        """Return a identifier to use as a key in a map of transactions."""
        return frozenset(self.package_names)

    def _reset_status(self):
        """Reset the current status progress."""
        self.status_string = ''
        self.percentage = 0
        self.stderr = None

    def install(self, skip_recommends=False, force_configuration=None,
                reinstall=False, force_missing_configuration=False):
        """Run an apt-get transaction to install given packages.

        If force_configuration is set to 'new', dpkg options will be enabled to
        make it force overwrite (without prompts) new configuration in place of
        old configuration (with a backup). This is useful when writing
        migration scripts in FreedomBox to handle the upgrades when
        unattended-upgrades refuse to upgrade a package due to configuration
        prompts.

        If force_configuration is set to 'old', dpkg options will be enabled to
        make it keep the old configuration (without prompts). This is useful
        when the Debian packages introduce new configuration with just
        cosmetics (such as updates to comments) and keeping the old
        configuration has same semantics.

        If force_configuration is None, no special options are passed to
        apt/dpkg for configuration file behavior.

        If reinstall is True, packages will be reinstalled, even if they are
        the latest version.

        If force_missing_configuration is True, any configuration files that
        have been removed after the first package has been installed will be
        restored.

        """
        try:
            privileged.update()
            kwargs = {
                'app_id': self.app_id,
                'packages': self.package_names,
                'skip_recommends': skip_recommends,
                'force_configuration': force_configuration,
                'reinstall': reinstall,
                'force_missing_configuration': force_missing_configuration
            }
            privileged.install(**kwargs)
        except Exception as exception:
            logger.exception('Error installing package: %s', exception)
            raise

    def uninstall(self, purge):
        """Run an apt-get transaction to uninstall given packages."""
        try:
            privileged.remove(app_id=self.app_id, packages=self.package_names,
                              purge=purge)
        except Exception as exception:
            logger.exception('Error uninstalling package: %s', exception)
            raise

    def refresh_package_lists(self):
        """Refresh apt package lists."""
        try:
            privileged.update()
        except Exception as exception:
            logger.exception('Error updating package lists: %s', exception)
            raise

    def _parse_progress(self, line):
        """Parse the apt-get process output line.

        See README.progress-reporting in apt source code.
        """
        parts = line.split(':')
        if len(parts) < 4:
            return

        status_map = {
            'pmstatus':
                gettext_lazy('installing'),
            'dlstatus':
                gettext_lazy('downloading'),
            'media-change':
                gettext_lazy('media change'),
            'pmconffile':
                format_lazy(gettext_lazy('configuration file: {file}'),
                            file=parts[1]),
        }
        self.status_string = status_map.get(parts[0], '')
        self.percentage = int(float(parts[2]))


def install(package_names, skip_recommends=False, force_configuration=None,
            reinstall=False, force_missing_configuration=False):
    """Install a set of packages marking progress."""
    try:
        operation = operation_module.Operation.get_operation()
    except AttributeError:
        raise RuntimeError(
            'install() must be called from within an operation.')

    if not operation.thread_data.get('allow_install', True):
        # Raise error if packages are not already installed.
        cache = apt.Cache()
        for package_name in package_names:
            if not cache[package_name].is_installed:
                raise PackageNotInstalledError(package_name)

        return

    start_time = time.time()
    while is_package_manager_busy():
        if time.time() - start_time >= 24 * 3600:  # One day
            raise PackageException(_('Timeout waiting for package manager'))

        time.sleep(3)  # seconds

    logger.info('Running install for app - %s, packages - %s',
                operation.app_id, package_names)

    from . import package
    transaction = package.Transaction(operation.app_id, package_names)
    operation.thread_data['transaction'] = transaction
    transaction.install(skip_recommends, force_configuration, reinstall,
                        force_missing_configuration)
    mark_known(package_names)


def uninstall(package_names, purge):
    """Uninstall a set of packages."""
    try:
        operation = operation_module.Operation.get_operation()
    except AttributeError:
        raise RuntimeError(
            'uninstall() must be called from within an operation.')

    start_time = time.time()
    while is_package_manager_busy():
        if time.time() - start_time >= 24 * 3600:  # One day
            raise PackageException(_('Timeout waiting for package manager'))

        time.sleep(3)  # seconds

    logger.info('Running uninstall for app - %s, packages - %s',
                operation.app_id, package_names)

    from . import package
    transaction = package.Transaction(operation.app_id, package_names)
    operation.thread_data['transaction'] = transaction
    transaction.uninstall(purge)
    unmark_known(package_names)


def is_package_manager_busy():
    """Return whether a package manager is running."""
    try:
        return privileged.is_package_manager_busy(_log_error=False)
    except Exception:
        return False


def refresh_package_lists():
    """To be run in case apt package lists are outdated."""
    transaction = Transaction(None, None)
    transaction.refresh_package_lists()


def filter_conffile_prompt_packages(packages):
    """Return a filtered info on packages that require conffile prompts.

    Information for each package includes: current_version, new_version and
    list of modified_conffiles.
    """
    return privileged.filter_conffile_packages(list(packages))


def packages_installed(candidates: list | tuple) -> list:
    """Check which candidates are installed on the system.

    :param candidates: A list of package names.
    :return: A list of installed Debian package names.
    """
    cache = apt.cache.Cache()
    installed_packages = []
    for package_name in candidates:
        try:
            package = cache[package_name]
            if package.is_installed:
                installed_packages.append(package_name)
        except KeyError:
            pass

    return installed_packages


def get_known() -> dict[str, dict]:
    """Return all the known packages and their versions.

    If a package is not known or has a version lower than the currently
    installed version, it means that the package has been installe or updated
    outside of FreedomBox. Some app, may use this information to rerun the
    setup on the app so that configuration is updated.
    """
    return kvstore.get_default('packages_known', {})


def mark_known(packages: list[str]):
    """Mark a given list of packages as known."""
    packages_known = get_known()
    cache = apt.Cache()
    for package_ in packages:
        try:
            cache_package = cache[package_]
        except KeyError:
            logger.warn('Package %s is not found when marking known', package_)
            continue

        if not cache_package.installed:
            logger.warn('Package %s is not installed when marking known',
                        package_)
            continue

        installed_version = cache_package.installed.version
        package_known = packages_known.setdefault(package_, {})
        package_known['version'] = installed_version

    kvstore.set('packages_known', packages_known)


def unmark_known(packages: list[str]):
    """Mark a give list of packages unknown."""
    packages_known = get_known()
    for package_ in packages:
        try:
            packages_known.pop(package_)
        except KeyError:
            pass

    kvstore.set('packages_known', packages_known)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Application manifest for Feather Wiki."""

from django.utils.translation import gettext_lazy as _

from .privileged import wiki_dir

clients = [{
    'name': _('Feather Wiki'),
    'platforms': [{
        'type': 'web',
        'url': '/featherwiki/'
    }]
}]

backup = {'data': {'directories': [str(wiki_dir)]}}

tags = [_('Wiki'), _('Note taking'), _('Website'), _('Quine'), _('Non-Debian')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for samba app.
"""

import random
import string
import subprocess
import urllib

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.samba]


class TestSambaApp(functional.BaseAppTests):
    app_name = 'samba'
    has_service = True
    has_web = False

    @pytest.fixture(scope='class', autouse=True)
    def fixture_setup(self, session_browser):
        """Setup the app."""
        functional.login(session_browser)
        functional.networks_set_firewall_zone(session_browser, 'internal')

    @pytest.mark.parametrize('share_type', ['open', 'group', 'home'])
    def test_enable_disable_samba_share(self, session_browser, share_type):
        """Test enabling and disabling Samba share."""
        _set_share(session_browser, share_type, status='enabled')

        _assert_share_is_writable(share_type)
        if share_type == 'open':
            _assert_share_is_writable(share_type, as_guest=True)
        else:
            _assert_share_is_not_accessible(share_type, as_guest=True)

        _set_share(session_browser, share_type, status='disabled')
        _assert_share_is_not_available(share_type)


def _set_share(browser, share_type, status='enabled'):
    """Enable or disable samba share."""
    disk_name = 'disk'
    share_row_id = 'samba-share-{0}-{1}'.format(disk_name, share_type)

    functional.nav_to_module(browser, 'samba')
    share = browser.find_by_id(share_row_id)
    share_btn = share.find_by_css('.share-status').find_by_tag('button').first

    if status == 'enabled' and share_btn['value'] == 'enable':
        functional.click_and_wait(browser, share_btn)
    elif status == 'disabled' and share_btn['value'] == 'disable':
        functional.click_and_wait(browser, share_btn)


def _write_to_share(share_type, as_guest=False):
    """Write to the samba share, return output messages as string."""
    disk_name = 'disk'
    default_url = functional.config['DEFAULT']['url']
    if share_type == 'open':
        share_name = disk_name
    else:
        share_name = '{0}_{1}'.format(disk_name, share_type)
    hostname = urllib.parse.urlparse(default_url).hostname
    servicename = '\\\\{0}\\{1}'.format(hostname, share_name)
    directory = '_plinth-test_{0}'.format(''.join(
        random.SystemRandom().choices(string.ascii_letters, k=8)))
    port = functional.config['DEFAULT']['samba_port']

    smb_command = ['smbclient', '-W', 'WORKGROUP', '-p', port]
    if as_guest:
        smb_command += ['-N']
    else:
        smb_command += [
            '-U', '{0}%{1}'.format(functional.config['DEFAULT']['username'],
                                   functional.config['DEFAULT']['password'])
        ]
    smb_command += [
        servicename, '-c', 'mkdir {0}; rmdir {0}'.format(directory)
    ]

    return subprocess.check_output(smb_command).decode()


def _assert_share_is_writable(share_type, as_guest=False):
    """Assert that samba share is writable."""
    output = _write_to_share(share_type, as_guest)

    assert not output, output


def _assert_share_is_not_accessible(share_type, as_guest=False):
    """Assert that samba share is not accessible."""
    try:
        _write_to_share(share_type, as_guest)
    except subprocess.CalledProcessError as err:
        err_output = err.output.decode()
        assert 'NT_STATUS_ACCESS_DENIED' in err_output, err_output
    else:
        assert False, 'Can access the share.'


def _assert_share_is_not_available(share_type):
    """Assert that samba share is not available."""
    try:
        _write_to_share(share_type)
    except subprocess.CalledProcessError as err:
        err_output = err.output.decode()
        assert 'NT_STATUS_BAD_NETWORK_NAME' in err_output, err_output
    else:
        assert False, 'Can access the share.'



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure Gitweb."""

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import frontpage, menu
from plinth.config import DropinConfigs
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.modules.users.components import UsersAndGroups
from plinth.package import Packages

from . import manifest, privileged
from .forms import is_repo_url
from .manifest import GIT_REPO_PATH

_description = [
    _('Git is a distributed version-control system for tracking changes in '
      'source code during software development. Gitweb provides a web '
      'interface to Git repositories. You can browse history and content of '
      'source code, use search to find relevant commits and code. '
      'You can also clone repositories and upload code changes with a '
      'command-line Git client or with multiple available graphical clients. '
      'And you can share your code with people around the world.'),
    _('To learn more on how to use Git visit '
      '<a href="https://git-scm.com/docs/gittutorial">Git tutorial</a>.')
]


class GitwebApp(app_module.App):
    """FreedomBox app for Gitweb."""

    app_id = 'gitweb'

    _version = 4

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        groups = {'git-access': _('Read-write access to Git repositories')}

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('Gitweb'), icon_filename='gitweb',
                               description=_description, manual_page='GitWeb',
                               clients=manifest.clients, tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-gitweb', info.name, info.icon_filename,
                              info.tags, 'gitweb:index',
                              parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut('shortcut-gitweb', info.name,
                                      icon=info.icon_filename, url='/gitweb/',
                                      clients=info.clients, tags=info.tags,
                                      login_required=True,
                                      allowed_groups=list(groups))
        self.add(shortcut)

        packages = Packages('packages-gitweb', ['gitweb', 'highlight'])
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-gitweb', [
            '/etc/gitweb-freedombox.conf',
            '/etc/apache2/conf-available/gitweb-freedombox.conf',
            '/etc/apache2/conf-available/gitweb-freedombox-auth.conf'
        ])
        self.add(dropin_configs)

        firewall = Firewall('firewall-gitweb', info.name,
                            ports=['http', 'https'], is_external=True)
        self.add(firewall)

        webserver = Webserver('webserver-gitweb', 'gitweb-freedombox',
                              urls=['https://{host}/gitweb/'])
        self.add(webserver)

        self.auth_webserver = GitwebWebserverAuth('webserver-gitweb-auth',
                                                  'gitweb-freedombox-auth')
        self.add(self.auth_webserver)

        users_and_groups = UsersAndGroups('users-and-groups-gitweb',
                                          groups=groups)
        self.add(users_and_groups)

        backup_restore = GitwebBackupRestore('backup-restore-gitweb',
                                             **manifest.backup)
        self.add(backup_restore)

    def post_init(self):
        """Perform post initialization operations."""
        if not self.needs_setup():
            self.update_service_access()

    def set_shortcut_login_required(self, login_required):
        """Change the login_required property of shortcut."""
        self.get_component('shortcut-gitweb').login_required = login_required

    def update_service_access(self):
        """Update the frontpage shortcut and webserver auth requirement."""
        repos = get_repo_list()
        if have_public_repos(repos):
            self._enable_public_access()
        else:
            self._disable_public_access()

    def _enable_public_access(self):
        """Allow Gitweb app to be accessed by anyone with access."""
        if self.is_enabled():
            self.auth_webserver.disable()

        self.set_shortcut_login_required(False)

    def _disable_public_access(self):
        """Allow Gitweb app to be accessed by logged-in users only."""
        if self.is_enabled():
            self.auth_webserver.enable()

        self.set_shortcut_login_required(True)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        if not old_version:
            self.enable()

    def uninstall(self):
        """De-configure and uninstall the app."""
        super().uninstall()
        privileged.uninstall()


class GitwebWebserverAuth(Webserver):
    """Component to handle Gitweb authentication webserver configuration."""

    is_leader = False

    def enable(self):
        """Enable apache configuration only if no public repository exists."""
        repos = get_repo_list()
        if not have_public_repos(repos):
            super().enable()

    def set_enabled(self, enabled):
        """Update the internal enabled state of the component.

        Actually do nothing. This method is called when is_leader = False to
        notify the enabled/disabled state of the app. Enabled/disabled state of
        this component is actually decided by the webserver configuration. It
        is expected that is_enabled is not called on this component to
        determine if app is enabled or not.
        """
        pass


class GitwebBackupRestore(BackupRestore):
    """Component to handle backup/restore for Gitweb."""

    def restore_post(self, packet):
        """Update access after restoration of backups."""
        super().restore_post(packet)
        self.app.update_service_access()


def have_public_repos(repos):
    """Check for public repositories."""
    return any((repo['access'] == 'public' for repo in repos))


def create_repo(repo, repo_description, owner, is_private):
    """Create a new repository or clone a remote repository."""
    kwargs = {
        'url': None,
        'name': None,
        'description': repo_description,
        'owner': owner,
        'is_private': is_private
    }

    if is_repo_url(repo):
        kwargs['url'] = repo
        # create a repo directory and set correct access rights
        privileged.create_repo(prepare_only=True, **kwargs)
        # start cloning in background
        privileged.create_repo(skip_prepare=True, _run_in_background=True,
                               **kwargs)
    else:
        kwargs['name'] = repo
        privileged.create_repo(**kwargs)


def get_repo_list():
    """List all git repositories."""
    repos = []
    if GIT_REPO_PATH.exists():
        for repo in GIT_REPO_PATH.iterdir():
            if not repo.name.endswith('.git') or repo.name.startswith('.'):
                continue

            repo_info = {}
            repo_info['name'] = repo.name[:-4]

            private_file = repo / 'private'
            if private_file.exists():
                repo_info['access'] = 'private'
            else:
                repo_info['access'] = 'public'

            progress_file = repo / 'clone_progress'
            if progress_file.exists():
                repo_info['clone_progress'] = progress_file.read_text()

            repos.append(repo_info)

    return sorted(repos, key=lambda repo: repo['name'])


def repo_info(repo):
    """Get information about repository."""
    info = privileged.repo_info(repo)
    if info['access'] == 'private':
        info['is_private'] = True
    else:
        info['is_private'] = False
    del info['access']

    return info


def edit_repo(form_initial, form_cleaned):
    """Edit repository data."""
    repo = form_initial['name']

    if form_cleaned['name'] != repo:
        privileged.rename_repo(repo, form_cleaned['name'])
        repo = form_cleaned['name']

    if form_cleaned['description'] != form_initial['description']:
        privileged.set_repo_description(repo, form_cleaned['description'])

    if form_cleaned['owner'] != form_initial['owner']:
        privileged.set_repo_owner(repo, form_cleaned['owner'])

    if form_cleaned['is_private'] != form_initial['is_private']:
        if form_cleaned['is_private']:
            privileged.set_repo_access(repo, 'private')
        else:
            privileged.set_repo_access(repo, 'public')

    if form_cleaned['default_branch'] != form_initial['default_branch']:
        privileged.set_default_branch(repo, form_cleaned['default_branch'])



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the plinth api for android app.
"""

from django.urls import re_path
from stronghold.decorators import public

from plinth.modules.api import views

urlpatterns = [
    re_path(r'^api/(?P<version>[0-9]+)/shortcuts/?$', public(views.shortcuts)),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Forms for security module
"""

from django import forms
from django.utils.translation import gettext_lazy as _


class SecurityForm(forms.Form):
    """Security configuration form"""
    fail2ban_enabled = forms.BooleanField(
        label=_('Fail2Ban (recommended)'), required=False,
        help_text=_('When this option is enabled, Fail2Ban will limit '
                    'brute force break-in attempts to the SSH server and '
                    'other enabled password protected internet-services.'))



# SPDX-License-Identifier: AGPL-3.0-or-later

import pathlib

from django.utils.translation import gettext_lazy as _

GIT_REPO_PATH = pathlib.Path('/var/lib/git')
REPO_DIR_OWNER = 'www-data'

clients = [
    {
        'name': _('Gitweb'),
        'platforms': [{
            'type': 'web',
            'url': '/gitweb/'
        }]
    },
    {
        'name':
            _('Git'),
        'platforms': [{
            'type': 'download',
            'os': 'gnu-linux',
            'url': 'https://git-scm.com/download/linux'
        }, {
            'type': 'download',
            'os': 'macos',
            'url': 'https://git-scm.com/download/mac'
        }, {
            'type': 'download',
            'os': 'windows',
            'url': 'https://git-scm.com/download/windows'
        }]
    },
]

backup = {'data': {'directories': [str(GIT_REPO_PATH)]}}

tags = [_('Git hosting'), _('Version control'), _('Developer tool')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for searx app.
"""

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.searx]


@pytest.fixture(scope='module', autouse=True)
def fixture_background(session_browser):
    """Login and install the app."""
    functional.login(session_browser)
    functional.install(session_browser, 'searx')
    yield
    functional.login(session_browser)
    functional.app_disable(session_browser, 'searx')


def test_enable_disable(session_browser):
    """Test enabling the app."""
    functional.app_disable(session_browser, 'searx')

    functional.app_enable(session_browser, 'searx')
    assert functional.is_available(session_browser, 'searx')
    _is_search_form_visible(session_browser)

    functional.app_disable(session_browser, 'searx')
    assert not functional.is_available(session_browser, 'searx')


@pytest.mark.backups
def test_backup_restore(session_browser):
    """Test backup and restore."""
    functional.app_enable(session_browser, 'searx')
    functional.backup_create(session_browser, 'searx', 'test_searx')
    functional.backup_restore(session_browser, 'searx', 'test_searx')
    assert functional.is_available(session_browser, 'searx')


def test_public_access(session_browser):
    """Test enabling public access."""
    functional.app_enable(session_browser, 'searx')

    # Enable public access
    _enable_public_access(session_browser)
    functional.logout(session_browser)
    assert functional.is_visible_on_front_page(session_browser, 'searx')
    assert functional.is_available(session_browser, 'searx')

    # Disable public access
    functional.login(session_browser)
    _disable_public_access(session_browser)
    functional.logout(session_browser)
    assert not functional.is_visible_on_front_page(session_browser, 'searx')
    assert not functional.is_available(session_browser, 'searx')


def test_preserve_public_access_setting(session_browser):
    """Test that public access setting is preserved when disabling and
    re-enabling the app."""
    functional.login(session_browser)
    functional.app_enable(session_browser, 'searx')
    _enable_public_access(session_browser)

    functional.app_disable(session_browser, 'searx')
    functional.app_enable(session_browser, 'searx')
    functional.logout(session_browser)

    assert functional.is_visible_on_front_page(session_browser, 'searx')
    assert functional.is_available(session_browser, 'searx')


def _enable_public_access(browser):
    """Enable Public Access in SearX"""
    functional.nav_to_module(browser, 'searx')
    browser.find_by_id('id_public_access').check()
    functional.submit(browser, form_class='form-configuration')


def _disable_public_access(browser):
    """Enable Public Access in SearX"""
    functional.nav_to_module(browser, 'searx')
    browser.find_by_id('id_public_access').uncheck()
    functional.submit(browser, form_class='form-configuration')


def _is_search_form_visible(browser):
    """Checks whether the search box is shown in the Searx web interface."""
    searx_app_url = functional.config['DEFAULT']['url'] + '/searx'
    browser.visit(searx_app_url)
    assert browser.find_by_id("search_form")



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for bind app.
"""

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.system, pytest.mark.bind]


class TestBindApp(functional.BaseAppTests):
    app_name = 'bind'
    has_service = True
    has_web = False

    def test_set_forwarders(self, session_browser):
        """Test setting forwarders."""
        functional.app_enable(session_browser, 'bind')
        functional.set_forwarders(session_browser, '1.1.1.1')

        functional.set_forwarders(session_browser, '1.1.1.1 1.0.0.1')
        assert functional.get_forwarders(session_browser) == '1.1.1.1 1.0.0.1'

    @pytest.mark.backups
    def test_backup_restore(self, session_browser):
        """Test backup and restore."""
        functional.app_enable(session_browser, 'bind')
        functional.set_forwarders(session_browser, '1.1.1.1')
        functional.backup_create(session_browser, 'bind', 'test_bind')

        functional.set_forwarders(session_browser, '1.0.0.1')

        functional.backup_restore(session_browser, 'bind', 'test_bind')
        assert functional.get_forwarders(session_browser) == '1.1.1.1'



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for quassel app.
"""

import pytest

from plinth.tests.functional import BaseAppTests

pytestmark = [pytest.mark.apps, pytest.mark.quassel]


class TestQuasselApp(BaseAppTests):
    app_name = 'quassel'
    has_service = True
    has_web = False

    # TODO: Improve test_backup_restore to actually check that data
    # configured servers is restored.



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for key/value store.
"""

import json
import os
import pathlib
import subprocess
from unittest.mock import Mock, call, patch

import pytest

from plinth.action_utils import (get_addresses, get_hostname,
                                 is_systemd_running, move_uploaded_file, run,
                                 run_as_user, service_action, service_disable,
                                 service_enable, service_is_enabled,
                                 service_is_running, service_reload,
                                 service_restart, service_start, service_stop,
                                 service_try_reload_or_restart,
                                 service_try_restart, service_unmask, umask)

UNKNOWN = 'unknowndeamon.service'
UNKNOWN_SOCKET = 'unknowndeamon.socket'

systemctl_path = pathlib.Path('/usr/bin/systemctl')
systemd_installed = pytest.mark.skipif(not systemctl_path.exists(),
                                       reason='systemd not available')

ip_path = pathlib.Path('/usr/bin/ip')
ip_installed = pytest.mark.skipif(not ip_path.exists(),
                                  reason='ip command not available')


@patch('os.path.exists')
def test_is_systemd_running(mock):
    """Trivial white box test for a trivial implementation."""
    mock.return_value = True
    assert is_systemd_running()
    mock.return_value = False
    assert not is_systemd_running()


@systemd_installed
def test_service_checks():
    """Test basic checks on status of an arbitrary service."""
    assert not service_is_running(UNKNOWN)
    assert not service_is_enabled(UNKNOWN)

    # expected is best if: generic. Alternatives: systemd-sysctl, logrotate
    expected = 'networking'
    if not service_is_running(expected):
        pytest.skip(f'Needs service {expected} running.')

    assert service_is_enabled(expected)


@pytest.mark.usefixtures('needs_root')
@systemd_installed
def test_service_enable_and_disable():
    """Test enabling and disabling of an arbitrary service."""
    # service is best if: non-essential part of FreedomBox that restarts fast
    service = 'unattended-upgrades'
    if not service_is_enabled(service):
        reason = f'Needs service {service} enabled.'
        pytest.skip(reason)

    service_disable(service)
    assert not service_is_running(service)
    service_enable(service)
    assert service_is_running(service)

    # Ignore unknown services, don't fail:
    service_disable(UNKNOWN)
    service_enable(UNKNOWN)


@patch('plinth.action_utils.service_action')
@systemd_installed
def test_service_start(mock):
    """Test staring a service."""
    service_start(UNKNOWN)
    mock.mock_calls = [call(UNKNOWN, 'start', check=False)]

    mock.reset_mock()
    service_start(UNKNOWN, check=True)
    mock.mock_calls = [call(UNKNOWN, 'start', check=True)]


@patch('plinth.action_utils.service_action')
@systemd_installed
def test_service_stop(mock):
    """Test stopping a service."""
    service_stop(UNKNOWN)
    assert mock.mock_calls == [call(UNKNOWN, 'stop', check=False)]

    mock.reset_mock()
    service_stop(UNKNOWN, check=True)
    mock.mock_calls = [call(UNKNOWN, 'stop', check=True)]

    mock.reset_mock()
    service_stop(UNKNOWN_SOCKET)
    assert mock.mock_calls == [
        call(UNKNOWN_SOCKET, 'stop', check=False),
        call(UNKNOWN, 'stop', check=False)
    ]


@patch('plinth.action_utils.service_action')
@systemd_installed
def test_service_restart(mock):
    """Test restaring a service."""
    service_restart(UNKNOWN)
    assert mock.mock_calls == [call(UNKNOWN, 'restart', check=False)]

    mock.reset_mock()
    service_restart(UNKNOWN, check=True)
    mock.mock_calls = [call(UNKNOWN, 'restart', check=True)]

    mock.reset_mock()
    service_restart(UNKNOWN_SOCKET)
    assert mock.mock_calls == [call(UNKNOWN, 'stop', check=False)]


@patch('plinth.action_utils.service_action')
@systemd_installed
def test_service_try_restart(mock):
    """Test try-restaring a service."""
    service_try_restart(UNKNOWN)
    assert mock.mock_calls == [call(UNKNOWN, 'try-restart', check=False)]

    mock.reset_mock()
    service_try_restart(UNKNOWN, check=True)
    mock.mock_calls = [call(UNKNOWN, 'try-restart', check=True)]

    mock.reset_mock()
    service_restart(UNKNOWN_SOCKET)
    assert mock.mock_calls == [call(UNKNOWN, 'stop', check=False)]


@patch('plinth.action_utils.service_action')
@systemd_installed
def test_service_reload(mock):
    """Test reloading a service."""
    service_reload(UNKNOWN)
    assert mock.mock_calls == [call(UNKNOWN, 'reload', check=False)]

    mock.reset_mock()
    service_reload(UNKNOWN, check=True)
    mock.mock_calls = [call(UNKNOWN, 'reload', check=True)]

    mock.reset_mock()
    service_reload(UNKNOWN_SOCKET)
    assert mock.mock_calls == [call(UNKNOWN, 'reload', check=False)]


@patch('plinth.action_utils.service_action')
@systemd_installed
def test_service_try_reload_or_restart(mock):
    """Test try-reload-or-restart on a service."""
    service_try_reload_or_restart(UNKNOWN)
    assert mock.mock_calls == [
        call(UNKNOWN, 'try-reload-or-restart', check=False)
    ]

    mock.reset_mock()
    service_try_reload_or_restart(UNKNOWN, check=True)
    mock.mock_calls = [call(UNKNOWN, 'try-reload-or-restart', check=True)]

    mock.reset_mock()
    service_try_reload_or_restart(UNKNOWN_SOCKET)
    assert mock.mock_calls == [
        call(UNKNOWN, 'try-reload-or-restart', check=False)
    ]


@pytest.mark.usefixtures('needs_root')
@systemd_installed
def test_service_unmask():
    """Test unmasking of an arbitrary masked service."""

    def is_masked(service):
        process = subprocess.run([
            'systemctl', 'list-unit-files', '--output=json',
            service + '.service'
        ], stdout=subprocess.PIPE, check=False)
        output = json.loads(process.stdout)
        return output[0]['state'] == 'masked' if output else False

    # SERVICE is best if: part of FreedomBox, so we can mess with least risk.
    service = 'samba-ad-dc'
    if not is_masked(service):
        pytest.skip(f'Needs service {service} masked.')

    service_unmask(service)
    assert not is_masked(service)

    service_action(service, 'mask')
    assert is_masked(service)


def test_get_hostname():
    """get_hostname returns a string.

    In fact, the maximum length for a hostname is 253 characters, but
    anything longer than 80 is very suspicious, so we fail the test.

    To avoid error messages pass as hostnames we seek and fail if we find
    some unexpected characters.
    """
    hostname = get_hostname()
    assert hostname
    assert isinstance(hostname, str)
    assert len(hostname) < 80
    for char in ' ,:;!?=$%&@*+()[]{}<>"\'':
        assert char not in hostname


@ip_installed
def test_get_addresses():
    """Test that any FreedomBox has some addresses."""
    ips = get_addresses()
    assert len(ips) > 3  # min: ip, 2x'localhost', hostname
    for address in ips:
        assert address['kind'] in ('4', '6')


@pytest.fixture(name='upload_dir')
def fixture_update_dir(tmp_path):
    """Patch Django file upload directory."""
    tmp_path /= 'source'
    tmp_path.mkdir()

    import plinth.settings
    old_value = plinth.settings.FILE_UPLOAD_TEMP_DIR
    plinth.settings.FILE_UPLOAD_TEMP_DIR = tmp_path
    yield tmp_path
    plinth.settings.FILE_UPLOAD_TEMP_DIR = old_value


def test_move_uploaded_file(tmp_path, upload_dir):
    """Test moving Django uploaded file to destination directory."""
    tmp_path /= 'destination'
    tmp_path.mkdir()

    # Source file does not exist
    source = tmp_path / 'does-non-exist'
    destination = tmp_path / 'destination'
    destination_file_name = 'destination-file-name'
    with pytest.raises(FileNotFoundError):
        move_uploaded_file(source, destination, destination_file_name)

    # Source is not a file
    source = tmp_path / 'source-dir'
    source.mkdir()
    with pytest.raises(ValueError, match='Source is not a file'):
        move_uploaded_file(source, destination, destination_file_name)

    # Source is not in expected temporary upload directory
    source = tmp_path / 'source-file'
    source.touch()
    with pytest.raises(
            ValueError,
            match='Uploaded file is not in expected temp directory'):
        move_uploaded_file(source, destination, destination_file_name)

    # Destination does not exist
    source = upload_dir / 'source-file'
    source.touch()
    with pytest.raises(ValueError, match='Destination is not a directory'):
        move_uploaded_file(source, destination, destination_file_name)

    # Destination is not a file
    destination.touch()
    with pytest.raises(ValueError, match='Destination is not a directory'):
        move_uploaded_file(source, destination, destination_file_name)

    # Destination file name is a multi-component path
    destination.unlink()
    destination.mkdir()
    destination_file_name = '../destination-file-name'
    with pytest.raises(ValueError, match='Invalid destination file name'):
        move_uploaded_file(source, destination, destination_file_name)

    # Destination file exists and override is not allowed
    destination_file_name = 'destination-file-exists'
    (destination / destination_file_name).touch()
    with pytest.raises(FileExistsError, match='Destination already exists'):
        move_uploaded_file(source, destination, destination_file_name)

    with pytest.raises(FileExistsError, match='Destination already exists'):
        move_uploaded_file(source, destination, destination_file_name,
                           allow_overwrite=False)

    # Successful move
    with patch('shutil.chown') as chown:
        destination_file = destination / destination_file_name
        destination_file.unlink()
        source.write_text('x-contents-1')
        move_uploaded_file(source, destination, destination_file_name)
        chown.mock_calls = [call(destination_file, 'root', 'root')]
        assert destination_file.stat().st_mode & 0o777 == 0o644
        assert destination_file.read_text() == 'x-contents-1'
        assert not source.exists()

        chown.reset_mock()
        source.write_text('x-contents-2')
        move_uploaded_file(source, destination, destination_file_name,
                           allow_overwrite=True, user='x-user',
                           group='x-group', permissions=0o600)
        chown.mock_calls = [call(destination_file, 'x-user', 'x-group')]
        assert destination_file.stat().st_mode & 0o777 == 0o600
        assert destination_file.read_text() == 'x-contents-2'
        assert not source.exists()


@patch('subprocess.run')
def test_run_as_user(subprocess_run):
    """Test running a command as another user works."""
    subprocess_run.return_value = 'test-return-value'
    return_value = run_as_user(['command', 'arg1', '--foo'],
                               username='foouser', stdout=subprocess.PIPE,
                               check=True)
    assert return_value == 'test-return-value'
    assert subprocess_run.mock_calls == [
        call(
            ['runuser', '--user', 'foouser', '--', 'command', 'arg1', '--foo'],
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
    ]


@patch('plinth.actions.thread_storage')
@patch('subprocess.run')
def test_run_capture(subprocess_run, thread_storage):
    """Test running a command with stdin/stdout capture works."""
    thread_storage.stdout = 'initial-stdout'
    thread_storage.stderr = 'initial-stderr'

    subprocess_run.return_value = Mock()
    subprocess_run.return_value.stdout = 'test-stdout'
    subprocess_run.return_value.stderr = 'test-stderr'

    return_value = run(['command', 'arg1', '--foo'], check=True)
    assert return_value == subprocess_run.return_value
    assert subprocess_run.mock_calls == [
        call(['command', 'arg1', '--foo'], stdout=subprocess.PIPE,
             stderr=subprocess.PIPE, check=True)
    ]
    assert thread_storage.stdout == 'initial-stdouttest-stdout'
    assert thread_storage.stderr == 'initial-stderrtest-stderr'


@patch('plinth.actions.thread_storage')
@patch('subprocess.run')
def test_run_no_capture(subprocess_run, thread_storage):
    """Test running a command without stdin/stdout capture works."""
    thread_storage.stdout = 'initial-stdout'
    thread_storage.stderr = 'initial-stderr'

    subprocess_run.return_value = Mock()
    subprocess_run.return_value.stdout = 'test-stdout'
    subprocess_run.return_value.stderr = 'test-stderr'

    return_value = run(['command', 'arg1', '--foo'], stdout=subprocess.PIPE,
                       stderr=subprocess.PIPE, check=True)
    assert return_value == subprocess_run.return_value
    assert subprocess_run.mock_calls == [
        call(['command', 'arg1', '--foo'], stdout=subprocess.PIPE,
             stderr=subprocess.PIPE, check=True)
    ]
    assert thread_storage.stdout == 'initial-stdout'
    assert thread_storage.stderr == 'initial-stderr'


@patch('plinth.actions.thread_storage', None)
@patch('subprocess.run')
def test_run_no_storage(subprocess_run):
    """Test running a command without thread storage."""
    subprocess_run.return_value = Mock()
    subprocess_run.return_value.stdout = 'test-stdout'
    subprocess_run.return_value.stderr = 'test-stderr'

    run(['command', 'arg1', '--foo'], check=True)


def test_umask(tmp_path):
    """Test that setting umask works."""

    def _asssert_umask(mask):
        """Assert the current umask."""
        old_umask = os.umask(0)
        assert old_umask == mask
        os.umask(old_umask)

    original_umask = os.umask(0o1)
    with umask(0o033):
        _asssert_umask(0o033)
        file1 = tmp_path / 'file1'
        file1.touch()
        assert file1.stat().st_mode & 0o033 == 0
        with umask(0o077):
            _asssert_umask(0o77)
            file2 = tmp_path / 'file2'
            file2.touch()
            assert file2.stat().st_mode & 0o077 == 0

        _asssert_umask(0o033)

    os.umask(original_umask)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Utilities for editing configuration files of Deluge.
"""

import copy
import json
import os
import pathlib
import re
import shutil
import tempfile

_JSON_FORMAT = {'indent': 4, 'sort_keys': True, 'ensure_ascii': False}


class Config:
    """Read or edit a Deluge configuration file."""

    def __init__(self, file_name):
        """Initialize the configuration object."""
        self.file_name = file_name
        self.file = pathlib.Path(self.file_name)
        self._version = None
        self.content = None
        self._original_content = None

    def load(self):
        """Parse the configuration file into memory."""
        text = self.file.read_text(encoding='utf-8')
        matches = re.match(r'^({[^}]*})(.*)$', text, re.DOTALL)
        if not matches:
            raise Exception('Unexpected file format.')

        try:
            self._version = json.loads(matches.group(1))
            self.content = json.loads(matches.group(2))
        except json.decoder.JSONDecodeError:
            raise Exception('Unable to parse JSON in file.')

        if self._version['format'] != 1:
            raise Exception('Version of the config file not understood')

        self._original_content = copy.deepcopy(self.content)

    def save(self):
        """Atomically save the modified configuration to file."""
        if self.content == self._original_content:
            return

        with tempfile.NamedTemporaryFile(dir=self.file.parent,
                                         delete=False) as new_file:
            new_file.write(json.dumps(self._version, **_JSON_FORMAT).encode())
            new_file.write(json.dumps(self.content, **_JSON_FORMAT).encode())
            new_file.flush()
            os.fsync(new_file.fileno())

        new_file_path = pathlib.Path(new_file.name)
        new_file_path.chmod(0o600)
        try:
            shutil.chown(str(new_file_path), 'debian-deluged',
                         'debian-deluged')
        except (PermissionError, LookupError):
            pass  # Not running as root, or deluge is not installed

        new_file_path.rename(self.file)

    def __enter__(self):
        """Enter the context."""
        self.load()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the context."""
        self.save()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Package holding all the privileged actions outside of apps."""

from .config import dropin_is_valid, dropin_link, dropin_unlink
from .container import (container_disable, container_enable,
                        container_is_enabled, container_setup,
                        container_uninstall)
from .packages import (filter_conffile_packages, install,
                       is_package_manager_busy, remove, update)
from .service import (disable, enable, get_logs, is_enabled, is_running, mask,
                      reload, restart, start, stop, systemd_set_default,
                      try_reload_or_restart, try_restart, unmask)

__all__ = [
    'filter_conffile_packages', 'install', 'is_package_manager_busy', 'remove',
    'update', 'systemd_set_default', 'disable', 'enable', 'is_enabled',
    'is_running', 'mask', 'reload', 'restart', 'start', 'stop',
    'try_reload_or_restart', 'try_restart', 'unmask', 'get_logs',
    'dropin_is_valid', 'dropin_link', 'dropin_unlink', 'container_disable',
    'container_enable', 'container_is_enabled', 'container_setup',
    'container_uninstall'
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for first boot wizard.
"""

import operator
import os
import sys

from django.urls import reverse
from django.utils.translation import gettext_noop

from plinth import app as app_module
from plinth import cfg
from plinth.signals import post_setup

first_boot_steps = [
    {
        'id': 'firstboot_welcome',
        'url': 'first_boot:welcome',
        'order': 0
    },
]

_all_first_boot_steps = None

_is_completed = None


class FirstBootApp(app_module.App):
    """FreedomBox app for First Boot."""

    app_id = 'first_boot'

    _version = 1

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               is_essential=True)
        self.add(info)

    def post_init(self):
        """Perform post initialization operations."""
        post_setup.connect(_clear_first_boot_steps)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)

        if not old_version:
            self._show_next_steps_notification()

        self.enable()

    def _show_next_steps_notification(self):
        """After first setup, show notification for next steps."""
        from plinth.notification import Notification
        title = gettext_noop('Setup complete! Next steps:')
        message = gettext_noop(
            'Initial setup has been completed. Perform the next steps to make '
            'your {box_name} operational.')
        data = {
            'app_name': 'translate:' + gettext_noop('Next steps'),
            'app_icon': 'fa-arrow-right',
            'box_name': 'translate:' + cfg.box_name
        }
        actions = [{
            'type': 'link',
            'class': 'primary',
            'text': gettext_noop('See next steps'),
            'url': 'first_boot:complete'
        }, {
            'type': 'dismiss'
        }]
        Notification.update_or_create(id='first-boot-complete',
                                      app_id='first_boot', severity='info',
                                      title=title, message=message,
                                      actions=actions, data=data,
                                      group='admin', dismissed=False)


def _clear_first_boot_steps(sender, module_name, **kwargs):
    """Flush the cache of first boot steps so it is recreated."""
    global _all_first_boot_steps
    _all_first_boot_steps = None


def is_firstboot_url(path):
    """Return whether a path is a firstboot step URL.

    :param path: path of url to be checked
    :return: true if its a first boot URL false otherwise
    """
    for step in _get_steps():
        if path.startswith(reverse(step['url'])):
            return True

    return False


def _get_steps():
    """Return list of all firstboot steps."""
    global _all_first_boot_steps
    if _all_first_boot_steps is not None:
        return _all_first_boot_steps

    steps = []
    for app in app_module.App.list():
        module = sys.modules[app.__module__]
        if getattr(module, 'first_boot_steps', None):
            if not app.needs_setup():
                steps.extend(module.first_boot_steps)

    _all_first_boot_steps = sorted(steps, key=operator.itemgetter('order'))
    return _all_first_boot_steps


def next_step():
    """Return the resolved next first boot step URL required to go to.

    If there are no more step remaining, return 'complete' page.
    """
    return next_step_or_none() or 'first_boot:complete'


def next_step_or_none():
    """Return the next first boot step required to run.

    If there are no more step remaining, return None.
    """
    from plinth import kvstore

    for step in _get_steps():
        done = kvstore.get_default(step['id'], 0)
        if not done:
            return step.get('url')


def mark_step_done(id):
    """Marks the status of a first boot step as done.

    :param id: id of the firstboot step
    """
    from plinth import kvstore

    kvstore.set(id, 1)
    if not next_step_or_none():
        set_completed()


def is_completed():
    """Return whether first boot process is completed."""
    from plinth import kvstore

    global _is_completed
    if _is_completed is None:
        _is_completed = kvstore.get_default('firstboot_completed', 0)

    return bool(_is_completed)


def set_completed():
    """Set the first boot process as completed."""
    from plinth import kvstore

    global _is_completed
    _is_completed = True
    kvstore.set('firstboot_completed', 1)


def get_secret_file_path():
    """Returns the path to the first boot wizard secret file."""
    return os.path.join(cfg.data_dir, 'firstboot-wizard-secret')


def firstboot_wizard_secret_exists():
    """Return whether a firstboot wizard secret exists."""
    secret_file = get_secret_file_path()
    return os.path.exists(secret_file) and os.path.getsize(secret_file) > 0



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Test the components provides by oidc app."""

from unittest.mock import patch

import pytest
from oauth2_provider.models import Application

from ..components import OpenIDConnect


@pytest.fixture(name='openid_connect')
def fixture_openid_connect():
    """Return an OpenIDConnect object."""
    return OpenIDConnect(
        'test-openidconnect', 'test-client', 'Test Client',
        ['https://testdomain.example/a', 'https://{domain}/b'])


def test_openidconnect_init():
    """Test initialization of openidconnect object."""
    component = OpenIDConnect('test-openidconnect', 'test-client',
                              'Test Client', ['https://testdomain.example/a'])
    assert component.component_id == 'test-openidconnect'
    assert component.client_id == 'test-client'
    assert component.name == 'Test Client'
    assert component.redirect_uris == ['https://testdomain.example/a']
    assert component.post_logout_redirect_uris == ''
    assert component.client_type == Application.CLIENT_CONFIDENTIAL
    assert component.authorization_grant_type == \
        Application.GRANT_AUTHORIZATION_CODE
    assert component.algorithm == Application.HS256_ALGORITHM
    assert not component.hash_client_secret
    assert not component.skip_authorization

    component = OpenIDConnect('test-openidconnect', 'test-client',
                              'Test Client', ['https://testdomain.example/a'],
                              skip_authorization=True)
    assert component.skip_authorization


@pytest.mark.django_db
def test_get_client_secret(openid_connect):
    """Test retrieving client secret."""
    with pytest.raises(Application.DoesNotExist):
        openid_connect.get_client_secret()

    openid_connect.setup(old_version=0)
    assert len(openid_connect.get_client_secret()) == 128


@pytest.mark.django_db
@patch('plinth.modules.names.components.DomainName.list_names')
def test_setup(list_names, openid_connect):
    """Test creating a DB object."""
    list_names.return_value = ('a.example', 'b.example')
    expected_redirect_uris = [
        'https://testdomain.example/a', 'https://a.example/b',
        'https://b.example/b', 'https://localhost/b'
    ]
    expected_origins = [
        'https://a.example', 'https://b.example', 'https://localhost'
    ]

    # Test creating fresh DB entry
    openid_connect.setup(old_version=0)
    obj = Application.objects.get(client_id=openid_connect.client_id)
    assert obj.client_id == openid_connect.client_id
    assert len(obj.client_secret) == 128
    assert not obj.user
    assert set(obj.redirect_uris.split(' ')) == set(expected_redirect_uris)
    assert obj.post_logout_redirect_uris == ''
    assert obj.client_type == Application.CLIENT_CONFIDENTIAL
    assert obj.authorization_grant_type == Application.GRANT_AUTHORIZATION_CODE
    assert not obj.hash_client_secret
    assert obj.name == 'Test Client'
    assert obj.algorithm == Application.HS256_ALGORITHM
    assert set(obj.allowed_origins.split(' ')) == set(expected_origins)
    assert not obj.skip_authorization

    # Test updating existing DB entry
    list_names.return_value = ('c.example', )
    expected_redirect_uris = ['https://c.example/c', 'https://localhost/c']
    expected_origins = ['https://c.example', 'https://localhost']
    openid_connect.redirect_uris = ['https://{domain}/c']
    openid_connect.allowed_origins = expected_origins
    openid_connect.post_logout_redirect_uris = 'a b'
    openid_connect.client_type = Application.CLIENT_PUBLIC
    openid_connect.authorization_grant_type = Application.GRANT_IMPLICIT
    openid_connect.hash_client_secret = True
    openid_connect.name = 'New name'
    openid_connect.algorithm = Application.RS256_ALGORITHM
    openid_connect.skip_authorization = True
    openid_connect.setup(old_version=0)
    obj = Application.objects.get(client_id=openid_connect.client_id)
    assert set(obj.redirect_uris.split(' ')) == set(expected_redirect_uris)
    assert set(obj.allowed_origins.split(' ')) == set(expected_origins)
    assert obj.post_logout_redirect_uris == 'a b'
    assert obj.client_type == Application.CLIENT_PUBLIC
    assert obj.authorization_grant_type == Application.GRANT_IMPLICIT
    assert obj.hash_client_secret
    assert obj.name == 'New name'
    assert obj.algorithm == Application.RS256_ALGORITHM
    assert obj.skip_authorization



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Python action utility functions.
"""

import logging
import os
import pathlib
import shutil
import subprocess
import tempfile
from contextlib import contextmanager
from typing import Generator

import augeas

from . import actions

logger = logging.getLogger(__name__)

# Flag on disk to indicate if freedombox package was held by
# plinth. This is a backup in case the process is interrupted and hold
# is not released.
apt_hold_flag = pathlib.Path('/var/lib/freedombox/package-held')


def is_systemd_running():
    """Return if we are running under systemd."""
    return os.path.exists('/run/systemd')


def systemd_get_default() -> str:
    """Return the default target that systemd will boot into."""
    process = run(['systemctl', 'get-default'], check=True)
    return process.stdout.decode().strip()


def systemd_set_default(target: str):
    """Set the default target that systemd will boot into."""
    run(['systemctl', 'set-default', target], check=True)


def service_daemon_reload():
    """Reload systemd to ensure that newer unit files are read."""
    run(['systemctl', 'daemon-reload'], check=True)


def service_is_running(servicename):
    """Return whether a service is currently running.

    Does not need to run as root.
    """
    try:
        run(['systemctl', 'status', servicename], check=True)
        return True
    except subprocess.CalledProcessError:
        # If a service is not running we get a status code != 0 and
        # thus a CalledProcessError
        return False


@contextmanager
def service_ensure_running(service_name):
    """Ensure a service is running and return to previous state."""
    starting_state = service_is_running(service_name)
    if not starting_state:
        service_enable(service_name)

    try:
        yield starting_state
    finally:
        if not starting_state:
            service_disable(service_name)


@contextmanager
def service_ensure_stopped(service_name):
    """Ensure a service is stopped and return to previous state."""
    starting_state = service_is_running(service_name)
    if starting_state:
        service_disable(service_name)

    try:
        yield starting_state
    finally:
        if starting_state:
            service_enable(service_name)


def service_is_enabled(service_name, strict_check=False):
    """Check if service is enabled in systemd.

    In some cases, after disabling a service, systemd puts it into a state
    called 'enabled-runtime' and returns a positive response to 'is-enabled'
    query. Until we understand better, a conservative work around is to pass
    strict=True to services effected by this behavior.

    """
    try:
        process = run(['systemctl', 'is-enabled', service_name], check=True)
        if not strict_check:
            return True

        return process.stdout.decode().strip() == 'enabled'
    except subprocess.CalledProcessError:
        return False


def service_enable(service_name: str, check: bool = False):
    """Enable and start a service in systemd."""
    run(['systemctl', 'enable', service_name], check=check)
    service_start(service_name, check=check)


def service_disable(service_name: str, check: bool = False):
    """Disable and stop service in systemd."""
    run(['systemctl', 'disable', service_name], check=check)
    try:
        service_stop(service_name, check=check)
    except subprocess.CalledProcessError:
        pass


def service_mask(service_name: str, check: bool = False):
    """Mask a service"""
    run(['systemctl', 'mask', service_name], check=check)


def service_unmask(service_name: str, check: bool = False):
    """Unmask a service"""
    run(['systemctl', 'unmask', service_name], check=check)


def service_start(service_name: str, check: bool = False):
    """Start a service with systemd."""
    service_action(service_name, 'start', check=check)

    # When starting a .socket unit, there is not need to start the .service
    # unit as it will be automatically started when a request is received on
    # the socket.


def _get_service_unit(socket_name: str) -> str:
    """Return the .service unit name for a .socket unit."""
    # Instead, may need to query the unit for associated .service file.
    base_name = socket_name.rpartition('.')[0]
    return f'{base_name}.service'


def service_stop(service_name: str, check: bool = False):
    """Stop a service with systemd."""
    service_action(service_name, 'stop', check=check)

    # When stopping a .socket unit, most of the time, we must also stop
    # .service unit. This frees up resources when disabling the app. It also
    # stops using resources that are being backed up.
    if service_name.endswith('.socket'):
        service_action(_get_service_unit(service_name), 'stop', check=check)


def service_restart(service_name: str, check: bool = False):
    """Restart a service with systemd."""
    if not service_name.endswith('.socket'):
        service_action(service_name, 'restart', check=check)
    else:
        # When restarting a .socket unit, most of the time, we actually want to
        # restart the corresponding .service unit. This reloads the
        # configuration changes as needed. To restart, all we need to do stop
        # the service. It will be automatically started again by .socket unit.
        service_action(_get_service_unit(service_name), 'stop', check=check)


def service_try_restart(service_name: str, check: bool = False):
    """Try to restart a service with systemd."""
    if not service_name.endswith('.socket'):
        service_action(service_name, 'try-restart', check=check)
    else:
        # When try-restarting a .socket unit, most of the time, we actually
        # want to restart the corresponding .service unit. This reloads the
        # configuration changes as needed. To restart, all we need to do stop
        # the service. It will be automatically started again by .socket unit.
        service_action(_get_service_unit(service_name), 'stop', check=check)


def service_reload(service_name: str, check: bool = False):
    """Reload a service with systemd."""
    if not service_name.endswith('.socket'):
        service_action(service_name, 'reload', check=check)
    else:
        # When reloading a .socket unit, most of the time, we actually want to
        # reload the corresponding .service unit. This reloads the
        # configuration changes as needed.
        service_action(_get_service_unit(service_name), 'reload', check=check)


def service_try_reload_or_restart(service_name: str, check: bool = False):
    """Reload a service if it supports reloading, otherwise restart.

    Do nothing if service is not running.
    """
    if not service_name.endswith('.socket'):
        service_action(service_name, 'try-reload-or-restart', check=check)
    else:
        # When reloading a .socket unit, most of the time, we actually want to
        # reload the corresponding .service unit. This reloads the
        # configuration changes as needed.
        service_action(_get_service_unit(service_name),
                       'try-reload-or-restart', check=check)


def service_reset_failed(service_name: str, check: bool = False):
    """Reset the 'failed' state of units."""
    service_action(service_name, 'reset-failed', check=check)


def service_get_logs(service_name: str) -> str:
    """Return the last lines of journal entries for a unit."""
    command = [
        'journalctl', '--no-pager', '--lines=200', '--unit', service_name
    ]
    process = run(command, check=False)
    return process.stdout.decode()


def service_show(service_name: str) -> dict[str, str]:
    """Return the status of the service in dictionary format."""
    command = ['systemctl', 'show', service_name]
    process = run(command, check=False)
    status = {}
    for line in process.stdout.decode().splitlines():
        parts = line.partition('=')
        status[parts[0]] = parts[2]

    return status


def service_action(service_name: str, action: str, check: bool = False):
    """Perform the given action on the service_name."""
    run(['systemctl', action, service_name], check=check)


def webserver_is_enabled(name, kind='config'):
    """Return whether a config/module/site is enabled in Apache."""
    if not shutil.which('a2query'):
        return False

    option_map = {'config': '-c', 'site': '-s', 'module': '-m'}
    try:
        # Don't print anything on the terminal
        run(['a2query', option_map[kind], name], check=True)
        return True
    except subprocess.CalledProcessError:
        return False


def webserver_enable(name, kind='config', apply_changes=True):
    """Enable a config/module/site in Apache.

    Restart/reload the webserver if apply_changes is True.  Return
    whether restart('restart'), reload('reload') or no action (None)
    is required.  If changes have been applied, then performed action
    is returned.
    """
    if webserver_is_enabled(name, kind) and kind == 'module':
        return

    command_map = {
        'config': 'a2enconf',
        'site': 'a2ensite',
        'module': 'a2enmod'
    }
    run([command_map[kind], name], check=True)

    action_required = 'restart' if kind == 'module' else 'reload'

    if apply_changes:
        if action_required == 'restart':
            service_restart('apache2')
        else:
            service_reload('apache2')

    return action_required


def webserver_disable(name, kind='config', apply_changes=True):
    """Disable config/module/site in Apache.

    Restart/reload the webserver if apply_changes is True.  Return
    whether restart('restart'), reload('reload') or no action (None)
    is required.  If changes have been applied, then performed action
    is returned.
    """
    if not webserver_is_enabled(name, kind):
        return

    command_map = {
        'config': 'a2disconf',
        'site': 'a2dissite',
        'module': 'a2dismod'
    }
    run([command_map[kind], name], check=True)

    action_required = 'restart' if kind == 'module' else 'reload'

    if apply_changes:
        if action_required == 'restart':
            service_restart('apache2')
        else:
            service_reload('apache2')

    return action_required


class WebserverChange:
    """Context to restart/reload Apache after configuration changes."""

    def __init__(self):
        """Initialize the context object state."""
        self.actions_required = set()

    def __enter__(self):
        """Return the context object so methods could be called on it."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Restart or reload the webserver.

        Don't suppress exceptions.  If an exception occurs
        restart/reload the webserver based on enable/disable
        operations done so far.
        """
        if 'restart' in self.actions_required:
            service_restart('apache2')
        elif 'reload' in self.actions_required:
            service_reload('apache2')

    def enable(self, name, kind='config'):
        """Enable a config/module/site in Apache.

        Don't apply the changes until the context is exited.
        """
        action_required = webserver_enable(name, kind, apply_changes=False)
        self.actions_required.add(action_required)

    def disable(self, name, kind='config'):
        """Disable a config/module/site in Apache.

        Don't apply the changes until the context is exited.
        """
        action_required = webserver_disable(name, kind, apply_changes=False)
        self.actions_required.add(action_required)


def get_addresses() -> list[dict[str, str | bool]]:
    """Return a list of IP addresses and hostnames."""
    addresses = get_ip_addresses()

    hostname = get_hostname()
    addresses.append({
        'kind': '4',
        'address': 'localhost',
        'numeric': False,
        'url_address': 'localhost'
    })
    addresses.append({
        'kind': '6',
        'address': 'localhost',
        'numeric': False,
        'url_address': 'localhost'
    })
    addresses.append({
        'kind': '4',
        'address': hostname,
        'numeric': False,
        'url_address': hostname
    })

    # When a hostname is resolved to IPv6 address, it may likely be link-local
    # address. Link local IPv6 addresses are valid only for a given link and
    # need to be scoped with interface name such as '%eth0' to work. Browsers
    # refused to implement support for link-local addresses (with zone IDs) in
    # URLs due to platform specific parsing rules and other implementation
    # difficulties. mod_auth_openidc does not support them either.
    # this correctly.

    return addresses


def get_ip_addresses() -> list[dict[str, str | bool]]:
    """Return a list of IP addresses assigned to the system."""
    addresses = []

    output = run(['ip', '-o', 'addr'], check=True).stdout
    for line in output.decode().splitlines():
        parts = line.split()
        address: dict[str, str | bool] = {
            'kind': '4' if parts[2] == 'inet' else '6',
            'address': parts[3].split('/')[0],
            'url_address': parts[3].split('/')[0],
            'numeric': True,
            'scope': parts[5],
            'interface': parts[1]
        }

        if address['kind'] == '6' and address['numeric']:
            address['url_address'] = '[{0}]'.format(address['address'])

        if address['scope'] != 'link':
            # Do not include link local addresses. Browsers refused to
            # implement support for link-local addresses (with zone IDs) in
            # URLs due to platform specific parsing rules and other
            # implementation difficulties. mod_auth_openidc does not support
            # them either.
            addresses.append(address)

    return addresses


def get_hostname():
    """Return the current hostname."""
    return run(['hostname'], check=True).stdout.decode().strip()


def dpkg_reconfigure(package, config):
    """Reconfigure package using debconf database override."""
    override_template = '''
Name: {package}/{key}
Template: {package}/{key}
Value: {value}
Owners: {package}
'''
    override_data = ''
    for key, value in config.items():
        override_data += override_template.format(package=package, key=key,
                                                  value=value)

    with tempfile.NamedTemporaryFile(mode='w', delete=False) as override_file:
        override_file.write(override_data)

    env = os.environ.copy()
    env['DEBCONF_DB_OVERRIDE'] = 'File{' + override_file.name + \
                                 ' readonly:true}'
    env['DEBIAN_FRONTEND'] = 'noninteractive'
    run(['dpkg-reconfigure', package], env=env, check=False)

    try:
        os.remove(override_file.name)
    except OSError:
        pass


def debconf_set_selections(presets):
    """Answer debconf questions before installing a package."""
    try:
        # Workaround Debian Bug #487300. In some situations, debconf complains
        # it can't find the question being answered even though it is supposed
        # to create a dummy question for it.
        run(['/usr/share/debconf/fix_db.pl'], check=True)
    except (FileNotFoundError, PermissionError):
        pass

    presets = '\n'.join(presets)
    run(['debconf-set-selections'], input=presets.encode(), check=True)


def is_disk_image():
    """Return whether the current machine is from a disk image.

    Two primary ways to install FreedomBox are:
    - Using FreedomBox image for various hardware platforms.
    - Installing packages on a Debian machine using apt.
    """
    return os.path.exists('/var/lib/freedombox/is-freedombox-disk-image')


def run_apt_command(arguments, enable_triggers: bool = False,
                    allow_freedombox_restart=False):
    """Run apt-get with provided arguments."""
    command = []
    if not allow_freedombox_restart:
        # Don't restart the freedombox web service. This configuration is only
        # used when apt command is invoked from freedombox web service itself
        # (such as during an app's installation/uninstallation).
        #
        # If this is not done, a freedombox web service restart is attempted.
        # needsrestart will wait until the restart is completed. apt command
        # will wait until needsrestart is completed. The restart mechanism in
        # service will wait until all currently running threads are completed.
        # One thread that has invoked this apt command will not finish as it
        # waits for apt command to finish. This results in a deadlock. Avoid
        # this by not attempting to restart freedombox web service when apt
        # command is invoked from freedombox web service.
        mount_path = '/etc/needrestart/conf.d/freedombox-self.conf'
        orig_path = f'/usr/share/freedombox{mount_path}'
        command = [
            'systemd-run', '--pipe',
            f'--property=BindReadOnlyPaths={orig_path}:{mount_path}'
        ]

    command += ['apt-get', '--assume-yes', '--quiet=2'] + arguments

    env = os.environ.copy()
    env['DEBIAN_FRONTEND'] = 'noninteractive'
    if not enable_triggers:
        env['FREEDOMBOX_INVOKED'] = 'true'
    process = run(command, stdin=subprocess.DEVNULL, env=env, check=False)
    return process.returncode


@contextmanager
def apt_hold(packages):
    """Prevent packages from being removed during apt operations.

    `apt-mark hold PACKAGES` accepts a list of packages. But if one of
    the package is missing from the apt repository, then it will fail
    to hold any of the listed packages. So it is necessary to try to
    hold each package by itself.

    Packages held by this context will be unheld when leaving the
    context. But if a package was already held beforehand, it will be
    ignored (and not unheld).

    """
    held_packages = []
    try:
        for package in packages:
            current_hold = run(['apt-mark', 'showhold', package],
                               check=True).stdout
            if not current_hold:
                process = run(['apt-mark', 'hold', package], check=False)
                if process.returncode == 0:  # success
                    held_packages.append(package)

        yield held_packages
    finally:
        for package in held_packages:
            run(['apt-mark', 'unhold', package], check=True)


@contextmanager
def apt_hold_freedombox():
    """Prevent freedombox package from being removed during apt operations."""
    current_hold = run(['apt-mark', 'showhold', 'freedombox'],
                       check=True).stdout
    try:
        if current_hold:
            # Package is already held, possibly by administrator.
            yield current_hold
        else:
            # Set the flag.
            apt_hold_flag.parent.mkdir(mode=0o755, parents=True, exist_ok=True)
            apt_hold_flag.touch(mode=0o660)
            yield run(['apt-mark', 'hold', 'freedombox'], check=True)
    finally:
        # Was the package held, either in this process or a previous one?
        if not current_hold or apt_hold_flag.exists():
            apt_unhold_freedombox()


def apt_unhold_freedombox():
    """Remove any hold on freedombox package, and clear flag."""
    run(['apt-mark', 'unhold', 'freedombox'], check=False)
    if apt_hold_flag.exists():
        apt_hold_flag.unlink()


def is_package_manager_busy():
    """Return whether package manager is busy.
    This command uses the `lsof` command to check whether the dpkg lock file
    is open which indicates that the package manager is busy"""
    LOCK_FILE = '/var/lib/dpkg/lock'
    try:
        run(['lsof', LOCK_FILE], check=True)
        return True
    except subprocess.CalledProcessError:
        return False


def podman_create(container_name: str, image_name: str, volume_name: str,
                  volume_path: str, volumes: dict[str, str] | None = None,
                  env: dict[str, str] | None = None,
                  binds_to: list[str] | None = None,
                  devices: dict[str, str] | None = None):
    """Remove and recreate a podman container."""
    service_stop(f'{volume_name}-volume.service')
    service_stop(container_name)

    # Data is kept
    run(['podman', 'volume', 'rm', '--force', volume_name], check=False)

    directory = pathlib.Path('/etc/containers/systemd')
    directory.mkdir(parents=True, exist_ok=True)

    # Fetch the image before creating the container. The systemd service for
    # the container won't timeout due to slow internet connectivity.
    run(['podman', 'image', 'pull', image_name], check=True)

    pathlib.Path(volume_path).mkdir(parents=True, exist_ok=True)
    # Create storage volume
    volume_file = directory / f'{volume_name}.volume'
    contents = f'''[Volume]
Device={volume_path}
Driver=local
VolumeName={volume_name}
Options=bind
'''
    volume_file.write_text(contents)

    service_file = directory / f'{container_name}.container'
    volume_lines = '\n'.join([
        f'Volume={source}:{dest}' for source, dest in (volumes or {}).items()
    ])
    env_lines = '\n'.join(
        [f'Environment={key}={value}' for key, value in (env or {}).items()])
    bind_lines = '\n'.join(f'BindsTo={service}\nAfter={service}'
                           for service in (binds_to or []))
    devices_lines = '\n'.join(f'AddDevice={source}:{dest}'
                              for source, dest in (devices or {}).items()
                              if pathlib.Path(source).exists())
    contents = f'''[Unit]
Requires={volume_name}-volume.service
After={volume_name}-volume.service
{bind_lines}

[Container]
AutoUpdate=registry
ContainerName=%N
{env_lines}
Image={image_name}
Network=host
{volume_lines}
{devices_lines}

[Service]
Restart=always

[Install]
WantedBy=default.target
'''
    service_file.write_text(contents)

    # Remove the fallback service file when upgrading from bookworm to trixie.
    # Re-running setup should be sufficient.
    _podman_create_fallback_service_file(container_name, image_name,
                                         volume_name, volume_path, volumes,
                                         env, binds_to, devices)

    service_daemon_reload()


def _podman_create_fallback_service_file(
        container_name: str, image_name: str, volume_name: str,
        volume_path: str, volumes: dict[str, str] | None = None,
        env: dict[str, str] | None = None, binds_to: list[str] | None = None,
        devices: dict[str, str] | None = None):
    """Create a systemd unit file if systemd generator is not available."""
    service_file = pathlib.Path(
        f'/etc/systemd/system/{container_name}.service')

    generator = '/usr/lib/systemd/system-generators/podman-system-generator'
    if pathlib.Path(generator).exists():
        # If systemd generator is present, during an upgrade, remove the
        # .service file (perhaps created when generator is not present).
        service_file.unlink(missing_ok=True)
        return

    service_file.parent.mkdir(parents=True, exist_ok=True)
    bind_lines = '\n'.join(f'BindsTo={service}\nAfter={service}'
                           for service in (binds_to or []))
    require_mounts_for = '\n'.join((f'RequiresMountsFor={host_path}'
                                    for host_path in (volumes or {})
                                    if host_path.startswith('/')))
    env_args = ' '.join(
        (f'--env {key}={value}' for key, value in (env or {}).items()))
    volume_args = ' '.join(
        (f'-v {host_path}:{container_path}'
         for host_path, container_path in (volumes or {}).items()))
    devices_args = ' '.join(
        (f'-d {host_path}:{container_path}'
         for host_path, container_path in (devices or {}).items()
         if pathlib.Path(host_path).exists()))

    # Similar to the file quadlet systemd generator produces but with volume
    # related commands merged.
    contents = f'''[Unit]
{bind_lines}
RequiresMountsFor=%t/containers
{require_mounts_for}

[Service]
Restart=always
Environment=PODMAN_SYSTEMD_UNIT=%n
KillMode=mixed
ExecStop=/usr/bin/podman rm -v -f -i --cidfile=%t/%N.cid
ExecStopPost=-/usr/bin/podman rm -v -f -i --cidfile=%t/%N.cid
Delegate=yes
Type=notify
NotifyAccess=all
SyslogIdentifier=%N
ExecStartPre=/usr/bin/rm -f %t/%N.cid
ExecStartPre=/usr/bin/podman volume rm --force {volume_name}
ExecStartPre=/usr/bin/podman volume create --driver=local --opt device={volume_path} --opt o=bind {volume_name}
ExecStart=/usr/bin/podman run --name=%N --cidfile=%t/%N.cid --replace --rm --cgroups=split --network=host --sdnotify=conmon --detach --label io.containers.autoupdate=registry {volume_args} {env_args} {devices_args} {image_name}

[Install]
WantedBy=default.target
'''  # noqa: E501
    service_file.write_text(contents, encoding='utf-8')
    service_daemon_reload()


def _podman_augeus(container_name: str):
    """Return an augues instance to edit container configuration file."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    container = f'/etc/containers/systemd/{container_name}.container'
    aug.transform('Systemd', container)
    aug.set('/augeas/context', '/files' + container)
    aug.load()
    return aug


def podman_is_enabled(container_name: str) -> bool:
    """Return whether the container to start on boot."""
    aug = _podman_augeus(container_name)
    aug = _podman_augeus(container_name)
    value = 'default.target'
    key = 'Install/WantedBy'
    return any(
        (aug.get(match_ + '/value') == value for match_ in aug.match(key)))


def podman_enable(container_name: str):
    """Enable container to start on boot."""
    aug = _podman_augeus(container_name)
    value = 'default.target'
    key = 'Install/WantedBy'
    found = any(
        (aug.get(match_ + '/value') == value for match_ in aug.match(key)))
    if not found:
        aug.set(f'{key}[last() +1]/value', value)
        aug.save()


def podman_disable(container_name: str):
    """Disable container to start on boot."""
    aug = _podman_augeus(container_name)
    aug.remove('Install/WantedBy')
    aug.save()


def podman_uninstall(container_name: str, volume_name: str, image_name: str,
                     volume_path: str):
    """Remove a podman container's components and systemd unit."""
    run(['podman', 'volume', 'rm', '--force', volume_name], check=True)
    run(['podman', 'image', 'rm', '--ignore', image_name], check=True)
    volume_file = pathlib.Path(
        '/etc/containers/systemd/') / f'{volume_name}.volume'
    volume_file.unlink(missing_ok=True)
    service_file = pathlib.Path(
        '/etc/containers/systemd/') / f'{container_name}.container'
    service_file.unlink(missing_ok=True)
    # Remove fallback service file
    service_file = pathlib.Path(
        '/etc/systemd/system/') / f'{container_name}.service'
    service_file.unlink(missing_ok=True)
    shutil.rmtree(volume_path, ignore_errors=True)
    service_daemon_reload()


def move_uploaded_file(source: str | pathlib.Path,
                       destination_dir: str | pathlib.Path,
                       destination_file_name: str,
                       allow_overwrite: bool = False, user: str = 'root',
                       group: str = 'root', permissions: int = 0o644):
    """Move an uploaded file from Django upload directory to a destination.

    Sets the expected ownership and permissions on the destination file. If
    possible, performs a simple rename operation. Otherwise, copies the file to
    the destination.

    The source must be regular file under the currently configured Django file
    upload directory. It must be a absolute path that can be verified to be
    under Django settings FILE_UPLOAD_TEMP_DIR.

    The destination_dir must be a directory. destination_file_name must be a
    simple file name without any other path components. When concatenated
    together, they specify the full destination path to move the file to.

    If allow_overwrite is set to False and destination file exists, an
    exception is raised.
    """
    from plinth import settings

    if isinstance(source, str):
        source = pathlib.Path(source)

    if isinstance(destination_dir, str):
        destination_dir = pathlib.Path(destination_dir)

    source = source.resolve(strict=True)
    destination_dir = destination_dir.resolve()

    # Verify source file
    if not source.is_file():
        raise ValueError('Source is not a file')

    tmp_dir = pathlib.Path(getattr(settings, 'FILE_UPLOAD_TEMP_DIR', '/tmp'))
    if not source.is_relative_to(tmp_dir):
        raise ValueError('Uploaded file is not in expected temp directory')

    # Verify destination directory
    if not destination_dir.is_dir():
        raise ValueError('Destination is not a directory')

    # Verify destination file name
    if len(pathlib.Path(destination_file_name).parts) != 1:
        raise ValueError('Invalid destination file name')

    destination = destination_dir / destination_file_name

    if destination.exists() and not allow_overwrite:
        raise FileExistsError('Destination already exists')

    # Move or copy
    shutil.move(source, destination)
    shutil.chown(destination, user, group)
    destination.chmod(permissions)


def run_as_user(command, username, **kwargs):
    """Run a command as another user.

    Uses 'runuser' which is similar to 'su'. Creates PAM session unlike
    setpriv. Sets real/effective uid/gid and resets the environment.
    """
    command = ['runuser', '--user', username, '--'] + command
    return run(command, **kwargs)


def run(command, **kwargs):
    """Run subprocess.run but capture stdout and stderr in thread storage."""
    collect_stdout = ('stdout' not in kwargs
                      and 'capture_output' not in kwargs)
    collect_stderr = ('stderr' not in kwargs
                      and 'capture_output' not in kwargs)

    if collect_stdout:
        kwargs['stdout'] = subprocess.PIPE

    if collect_stderr:
        kwargs['stderr'] = subprocess.PIPE

    try:
        process = subprocess.run(command, **kwargs)
        if collect_stdout and hasattr(actions.thread_storage, 'stdout'):
            actions.thread_storage.stdout += process.stdout

        if collect_stderr and hasattr(actions.thread_storage, 'stderr'):
            actions.thread_storage.stderr += process.stderr
    except subprocess.CalledProcessError as exception:
        if exception.stdout and hasattr(actions.thread_storage, 'stdout'):
            actions.thread_storage.stdout += exception.stdout

        if exception.stderr and hasattr(actions.thread_storage, 'stderr'):
            actions.thread_storage.stderr += exception.stderr

        raise exception

    return process


@contextmanager
def umask(mask) -> Generator:
    """Set the umask temporarily for a operation and then revert it."""
    old_umask = os.umask(mask)
    try:
        yield
    finally:
        os.umask(old_umask)



#!/usr/bin/python3
# SPDX-License-Identifier: AGPL-3.0-or-later

import argparse
import logging
import os
import sys
import threading

import systemd.daemon

from . import __version__
from . import app as app_module
from . import (cfg, frontpage, glib, log, menu, module_loader, setup,
               web_framework, web_server)

precedence_commandline_arguments = ["server_dir", "develop"]

logger = logging.getLogger(__name__)


def parse_arguments():
    """Parse command line arguments"""
    parser = argparse.ArgumentParser(
        description='Core functionality and web interface for FreedomBox',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    # TODO: server_dir is actually a url prefix; use a better variable name
    parser.add_argument('--server_dir', default=None,
                        help='web server path under which to serve')
    parser.add_argument(
        '--develop', action='store_true', default=None,
        help=('run Plinth *insecurely* from current folder; '
              'enable auto-reloading and debugging options'))
    parser.add_argument('--setup', default=False, nargs='*',
                        help='run setup tasks on all essential apps and exit')
    parser.add_argument(
        '--setup-no-install', default=False, nargs='*',
        help='run setup tasks without installing packages and exit')
    parser.add_argument('--list-dependencies', default=False, nargs='*',
                        help='list package dependencies for essential modules')
    parser.add_argument('--list-apps', default=False, nargs='*',
                        help='list apps')
    parser.add_argument('--version', action='store_true', default=None,
                        help='show version and exit')

    return parser.parse_args()


def run_setup_and_exit(app_ids, allow_install=True):
    """Run setup on all essential apps and exit."""
    error_code = 0
    try:
        setup.run_setup_on_apps(app_ids, allow_install)
    except Exception:
        error_code = 1

    sys.exit(error_code)


def list_dependencies(app_ids):
    """List dependencies for all essential apps and exit."""
    error_code = 0
    try:
        if app_ids:
            setup.list_dependencies(app_ids=app_ids)
        else:
            setup.list_dependencies(essential=True)
    except Exception as exception:
        logger.error('Error listing dependencies - %s', exception)
        error_code = 1

    sys.exit(error_code)


def list_apps(apps_type):
    """List all/essential/optional apps and exit."""
    for app in app_module.App.list():
        is_essential = app.info.is_essential
        if 'essential' in apps_type and not is_essential:
            continue

        if 'optional' in apps_type and is_essential:
            continue

        print(f'{app.app_id}')

    sys.exit()


def adapt_config(arguments):
    """Give commandline arguments precedence over config entries"""
    for argument_name in precedence_commandline_arguments:
        argument_value = getattr(arguments, argument_name)
        if argument_value is not None:
            setattr(cfg, argument_name, argument_value)


def on_web_server_stop():
    """Stop all other threads since web server is trying to exit."""
    setup.stop()
    glib.stop()


def run_post_init_and_setup():
    """Run post-init operations on the apps and setup operations."""
    app_module.apps_post_init()
    frontpage.add_custom_shortcuts()

    # Handle app version updates.
    setup.run_setup_on_startup()  # Long running, retrying

    # Handle packages that have been updated else where that need a re-run of
    # setup.
    setup.on_dpkg_invoked()

    # Handle packages that have a pending configuration file prompt.
    setup.on_package_cache_updated()


def main():
    """Initialize and start the application"""
    arguments = parse_arguments()

    cfg.read()
    if arguments.develop or os.getenv('FREEDOMBOX_DEVELOP', '') == '1':
        cfg.develop = True
        # Use the config in the current working directory
        cfg.read_file(cfg.get_develop_config_path())

    adapt_config(arguments)

    if arguments.version:
        print(f'FreedomBox {__version__}')
        sys.exit(0)

    if arguments.list_dependencies is not False:
        log.default_level = 'ERROR'
        module_loader.load_modules()
        app_module.apps_init()
        list_dependencies(arguments.list_dependencies)

    if arguments.list_apps is not False:
        log.default_level = 'ERROR'
        module_loader.load_modules()
        app_module.apps_init()
        list_apps(arguments.list_apps)

    log.init()

    web_framework.init()
    web_framework.post_init()

    logger.info('FreedomBox Service (Plinth) version - %s', __version__)
    for config_file in cfg.config_files:
        logger.info('Configuration loaded from file - %s', config_file)
    logger.info('Script prefix - %s', cfg.server_dir)

    module_loader.include_urls()

    menu.init()

    module_loader.load_modules()
    app_module.apps_init()

    if arguments.setup is not False:
        app_module.apps_post_init()
        run_setup_and_exit(arguments.setup, allow_install=True)

    if arguments.setup_no_install is not False:
        app_module.apps_post_init()
        run_setup_and_exit(arguments.setup_no_install, allow_install=False)

    threading.Thread(target=run_post_init_and_setup).start()

    glib.run()

    web_server.init()
    web_server.run(on_web_server_stop)

    # systemd will wait until notification to proceed with other processes. We
    # have service Type=notify.
    systemd.daemon.notify('READY=1')

    web_server.block()


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

backup = {'secrets': {'directories': ['/etc/coturn']}, 'services': ['coturn']}

tags = [_('Video conference'), _('STUN'), _('TURN')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Tests for config module.
"""

import os
import pathlib
from unittest.mock import Mock, patch

import pytest

from plinth import __main__ as plinth_main
from plinth import utils
from plinth.modules.apache import uws_directory_of_user, uws_url_of_user
from plinth.modules.config import (_home_page_scid2url, change_home_page,
                                   get_home_page, home_page_url2scid)


def test_homepage_mapping():
    """Basic tests for homepage functions."""
    func = home_page_url2scid
    assert func(None) == 'apache-default'
    assert func('/unknown/url') is None
    assert func('/plinth/') == 'freedombox'
    assert func('/plinth') == 'freedombox'
    assert func('plinth') == 'freedombox'
    assert func('/freedombox/') == 'freedombox'
    assert func('/freedombox') == 'freedombox'
    assert func('freedombox') == 'freedombox'
    assert func('/index.html') == 'apache-default'
    assert func('/~user') == 'uws-user'
    assert func('/~user/whatever/else') == 'uws-user'

    func = _home_page_scid2url
    assert func('freedombox') == '/freedombox/'
    assert func('apache-default') is None


def test_homepage_mapping_skip_ci():
    """Special tests for homepage functions."""
    try:
        user = os.getlogin()
    except OSError:
        # See msg383161 in https://bugs.python.org/issue40821
        reason = "Needs access to ~/ directory. " \
               + "CI sandboxed workspace doesn't provide it."
        pytest.skip(reason)
    uws_directory = uws_directory_of_user(user)
    uws_url = uws_url_of_user(user)
    uws_scid = home_page_url2scid(uws_url)

    # Check test's precondition:
    if os.path.exists(uws_directory):
        # Don't blindly remove a pre-existing directory. Just skip the test.
        reason = "UWS directory {} exists already.".format(uws_directory)
        pytest.skip(reason)

    # AC: Return scid if UWS directory exists:
    try:
        os.mkdir(uws_directory)
    except Exception:
        reason = "Needs access to ~/ directory. " \
               + "CI sandboxed workspace doesn't provide it."
        pytest.skip(reason)
    assert _home_page_scid2url(uws_scid) == uws_url

    # AC: Return None if it doesn't:
    os.rmdir(uws_directory)
    assert _home_page_scid2url(uws_scid) == '/freedombox/'


@patch(
    'plinth.frontpage.Shortcut.list',
    Mock(return_value=[
        Mock(url='url/for/' + id, component_id=id) for id in ('a', 'b')
    ]))
@pytest.mark.usefixtures('needs_root')
def test_homepage_field():
    """Test homepage changes.

    Test Cases:
        1) FreedomBox Homepage (default),
        2) Apache default,
        3) A user's website of an...
        3.1) unexisting user
        3.2) existing user without a page
        3.3) existing user page.
        4) A FreedomBox App.
        4.1) unknown app
        4.2) uninstalled app
        4.3) disabled app
        4.4) enabled app

    Note: If run on a pristine unconfigured FreedomBox, this test will leave
          the homepage default-configured. (Imperfect cleanup in such case).

    Note: We take fbx as website user because of our testing container.

    Pending: - Specific test cases to distinguish 4.1,2,3.
               Currently they share the same test case.
             - Search for another valid user apart from fbx.
    """
    user = 'test_' + utils.random_string(size=12)
    uws_directory = uws_directory_of_user(user)
    uws_url = uws_url_of_user(user)
    uws_scid = home_page_url2scid(uws_url)

    default_home_page = 'freedombox'
    original_home_page = get_home_page() or default_home_page
    change_home_page(default_home_page)  # Set to known value explicitly

    # AC: invalid changes fall back to default:
    for scid in ('uws-unexisting', uws_scid, 'missing_app'):
        change_home_page(scid)
        assert get_home_page() == default_home_page

    # AC: valid changes actually happen:
    pathlib.Path(uws_directory).mkdir(parents=True)
    for scid in ('b', 'a', uws_scid, 'apache-default', 'freedombox'):
        change_home_page(scid)
        assert get_home_page() == scid

    # cleanup:
    change_home_page(original_home_page)
    os.rmdir(uws_directory)
    assert get_home_page() == original_home_page


def test_locale_path():
    """
    Test that the 'locale' directory is in the same folder as __main__.py.
    This is required for detecting translated languages.
    """
    plinth_dir = os.path.dirname(plinth_main.__file__)
    locale_dir = os.path.join(plinth_dir, 'locale')
    assert os.path.isdir(locale_dir)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app to configure Shaarli.
"""

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import frontpage, menu
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.package import Packages

from . import manifest

_description = [
    _('Shaarli allows you to save and share bookmarks.'),
    _('Note that Shaarli only supports a single user account, which you will '
      'need to setup on the initial visit.'),
]


class ShaarliApp(app_module.App):
    """FreedomBox app for Shaarli."""

    app_id = 'shaarli'

    _version = 1

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('Shaarli'), icon_filename='shaarli',
                               description=_description, manual_page='Shaarli',
                               clients=manifest.clients, tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-shaarli', info.name, info.icon_filename,
                              info.tags, 'shaarli:index',
                              parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut('shortcut-shaarli', info.name,
                                      icon=info.icon_filename, url='/shaarli',
                                      clients=info.clients, tags=info.tags,
                                      login_required=True)
        self.add(shortcut)

        packages = Packages('packages-shaarli', ['shaarli'])
        self.add(packages)

        firewall = Firewall('firewall-shaarli', info.name,
                            ports=['http', 'https'], is_external=True)
        self.add(firewall)

        webserver = Webserver('webserver-shaarli', 'shaarli')
        self.add(webserver)

        backup_restore = BackupRestore('backup-restore-shaarli',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        if not old_version:
            self.enable()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Tor Proxy utility functions."""

import itertools

import augeas

from plinth import app as app_module
from plinth.daemon import app_is_running

from . import privileged

APT_SOURCES_URI_PATHS = ('/files/etc/apt/sources.list/*/uri',
                         '/files/etc/apt/sources.list.d/*/*/uri',
                         '/files/etc/apt/sources.list.d/*/*/URIs/*')
APT_TOR_PREFIX = 'tor+'


def get_status(initialized=True):
    """Return current Tor status."""
    status = privileged.get_status()

    app = app_module.App.get('torproxy')
    return {
        'enabled': app.is_enabled() if initialized else False,
        'is_running': app_is_running(app) if initialized else False,
        'use_upstream_bridges': status['use_upstream_bridges'],
        'upstream_bridges': status['upstream_bridges'],
        'apt_transport_tor_enabled': is_apt_transport_tor_enabled()
    }


def iter_apt_uris(aug):
    """Iterate over all the APT source URIs."""
    return itertools.chain.from_iterable(
        [aug.match(path) for path in APT_SOURCES_URI_PATHS])


def get_augeas(raise_exception=True):
    """Return an instance of Augeaus for processing APT configuration."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/Aptsources/lens', 'Aptsources.lns')
    aug.set('/augeas/load/Aptsources/incl[last() + 1]',
            '/etc/apt/sources.list')
    aug.set('/augeas/load/Aptsources/incl[last() + 1]',
            '/etc/apt/sources.list.d/*.list')
    aug.set('/augeas/load/Aptsources822/lens', 'Aptsources822.lns')
    aug.set('/augeas/load/Aptsources822/incl[last() + 1]',
            '/etc/apt/sources.list.d/*.sources')
    aug.load()

    # Check for any errors in parsing sources lists.
    if raise_exception:
        if aug.match('/augeas/files/etc/apt/sources.list/error') or \
           aug.match('/augeas/files/etc/apt/sources.list.d//error'):
            raise Exception('Error parsing sources list')

    return aug


def is_apt_transport_tor_enabled():
    """Return whether APT is set to download packages over Tor."""
    try:
        aug = get_augeas()
    except Exception:
        # If there was an error with parsing.
        return False

    for uri_path in iter_apt_uris(aug):
        uri = aug.get(uri_path)
        if not uri.startswith(APT_TOR_PREFIX) and \
           (uri.startswith('http://') or uri.startswith('https://')):
            return False

    return True



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Configures rspamd to handle incoming and outgoing spam.

See: http://www.postfix.org/MILTER_README.html
See: https://rspamd.com/doc/configuration/ucl.html

For testing DKIM signatures: https://www.mail-tester.com/
"""

import pathlib
import re

from plinth import action_utils
from plinth.actions import privileged
from plinth.modules.email import postfix

_milter_config = {
    'smtpd_milters': 'inet:127.0.0.1:11332',
    'non_smtpd_milters': 'inet:127.0.0.1:11332',
}


@privileged
def setup_spam():
    """Compile sieve filters and set rspamd/postfix configuration."""
    _compile_sieve()
    _setup_rspamd()
    postfix.set_config(_milter_config)


def _compile_sieve():
    """Compile all .sieve script to binary format for performance."""
    sieve_dirs = [
        '/etc/dovecot/freedombox-sieve-after/', '/etc/dovecot/freedombox-sieve'
    ]
    for sieve_dir in sieve_dirs:
        action_utils.run(['sievec', sieve_dir], check=True)


def _setup_rspamd():
    """Adjust configuration to include FreedomBox configuration files."""
    configs = [('milter_headers.conf', 'freedombox-milter-headers.conf'),
               ('redis.conf', 'freedombox-redis.conf'),
               ('logging.inc', 'freedombox-logging.inc'),
               ('dkim_signing.conf', 'freedombox-dkim-signing.conf')]
    base_path = pathlib.Path('/etc/rspamd/local.d')
    for orig_path, include_path in configs:
        _setup_local_include(base_path / orig_path, base_path / include_path)


def _setup_local_include(orig_path, include_path):
    """Adjust configuration to include a FreedomBox configuration file."""
    lines = []
    if orig_path.exists():
        lines = orig_path.read_text().splitlines()

    file_name = include_path.name
    for line in lines:
        if re.match(rf'\s*.include\(.*\)\s+".*/{file_name}"', line):
            return

    lines.append('.include(priority=2,duplicate=merge) '
                 f'"$LOCAL_CONFDIR/local.d/{file_name}"\n')
    orig_path.write_text('\n'.join(lines))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for Miniflux app.
"""

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.miniflux]

ADMIN_USERNAME = 'admin'
ADMIN_PASSWORD = 'str0ngp@$$word'
ADMIN_PASSWORD_NEW = 'str0ngERp@$$word'

CREDENTIALS = {'username': 'admin', 'password': ADMIN_PASSWORD}


class TestMinifluxApp(functional.BaseAppTests):
    """Class to customize basic app tests for Miniflux."""

    app_name = 'miniflux'
    has_service = True
    has_web = True

    @pytest.fixture(name='create_admin_user')
    def fixture_create_admin_user(self, session_browser):
        """Create an admin user for Miniflux."""
        functional.app_enable(session_browser, self.app_name)
        _create_admin_user(session_browser)

    def test_create_miniflux_admin_user(self, session_browser,
                                        create_admin_user):
        """Test creating an admin user."""
        _miniflux_login(session_browser)
        # Verify that this user can see admin settings
        functional.click_link_by_href(session_browser, '/miniflux/settings')

        assert not session_browser.links.find_by_href(
            '/miniflux/users').is_empty()

    def test_reset_miniflux_user_password(self, session_browser,
                                          create_admin_user):
        """Test Miniflux user password reset."""
        CREDENTIALS['password'] = ADMIN_PASSWORD_NEW
        _reset_user_password(session_browser)
        _miniflux_login(session_browser)
        assert not session_browser.links.find_by_href(
            '/miniflux/unread').is_empty()

    @pytest.mark.backups
    def test_backup_restore(self, session_browser, create_admin_user):
        """Test backup and restore of app data."""
        _subscribe(session_browser, 'https://planet.debian.org/atom.xml')
        super().test_backup_restore(session_browser)
        assert _is_subscribed(session_browser, 'Planet Debian')


def _fill_credentials_form(browser, href):
    """Fill the user credentials form in Miniflux app."""
    functional.nav_to_module(browser, 'miniflux')
    functional.click_link_by_href(browser,
                                  f'/freedombox/apps/miniflux/{href}/')

    browser.fill('miniflux-username', CREDENTIALS['username'])
    browser.fill('miniflux-password', CREDENTIALS['password'])
    browser.fill('miniflux-password_confirmation', CREDENTIALS['password'])
    functional.submit(browser, form_class='form-miniflux')


def _create_admin_user(browser):
    """Create Miniflux admin user."""
    _fill_credentials_form(browser, 'create-admin-user')
    _fill_credentials_form(browser, 'reset-user-password')


def _open_miniflux_app(browser):
    """Load the web interface of Miniflux."""
    functional.visit(browser, '/miniflux/')
    main = browser.find_by_id('main')
    functional.eventually(lambda: main.visible)


def _miniflux_logout(browser):
    """Attempt to log out of Miniflux app. Doesn't fail if not logged in."""
    _open_miniflux_app(browser)
    maybe_logout_button = browser.links.find_by_href('/miniflux/logout')
    if not maybe_logout_button.is_empty():
        functional.click_and_wait(browser, maybe_logout_button)


def _miniflux_submit(browser):
    """Perform the Submit action in Miniflux forms."""
    functional.submit(browser,
                      element=browser.find_by_css('button[type="submit"]'))


def _miniflux_login(browser):
    """Login to miniflux with the given credentials."""
    _open_miniflux_app(browser)
    _miniflux_logout(browser)
    browser.find_by_id('form-username').fill(CREDENTIALS['username'])
    browser.find_by_id('form-password').fill(CREDENTIALS['password'])
    _miniflux_submit(browser)


def _reset_user_password(browser):
    """Reset a Miniflux user's password from FreedomBox web interface."""
    _fill_credentials_form(browser, 'reset-user-password')


def _subscribe(browser, feed_url):
    """Subscribe to a feed in Miniflux."""
    _open_miniflux_app(browser)
    _miniflux_login(browser)
    functional.click_link_by_href(browser, '/miniflux/subscribe')
    browser.find_by_id('form-url').fill(feed_url)
    _miniflux_submit(browser)


def _is_subscribed(browser, feed_name):
    """Check if the user is subscribed to a feed."""
    _open_miniflux_app(browser)
    _miniflux_login(browser)
    functional.click_link_by_href(browser, '/miniflux/feeds')

    return browser.is_text_present(feed_name)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure sharing."""

import os
import pathlib
import re

import augeas

from plinth import action_utils
from plinth.actions import privileged

APACHE_CONFIGURATION = '/etc/apache2/conf-available/sharing-freedombox.conf'


@privileged
def setup():
    """Create an empty apache configuration file."""
    path = pathlib.Path(APACHE_CONFIGURATION)
    if not path.exists():
        path.touch()

    _migrate_old_style_auth()


def _migrate_old_style_auth():
    """Migration from auth_pubtkt to auth_openidc."""
    aug = load_augeas()
    if not aug.match('$conf//directive["TKTAuthToken"]'):
        return

    shares = _list(aug)
    for share in shares:
        _remove(aug, share['name'])

    for share in shares:
        _add(aug, share['name'], share['path'], share['groups'],
             share['is_public'])

    aug.save()
    action_utils.service_reload('apache2')


def load_augeas():
    """Initialize augeas for this app's configuration file."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/Httpd/lens', 'Httpd.lns')
    aug.set('/augeas/load/Httpd/incl[last() + 1]', APACHE_CONFIGURATION)
    aug.load()

    aug.defvar('conf', '/files' + APACHE_CONFIGURATION)

    return aug


@privileged
def add(name: str, path: str, groups: list[str], is_public: bool):
    """Add a share to Apache configuration."""

    if not os.path.exists(APACHE_CONFIGURATION):
        pathlib.Path(APACHE_CONFIGURATION).touch()

    aug = load_augeas()
    shares = _list(aug)
    if any([share for share in shares if share['name'] == name]):
        raise Exception('Share already present')

    _add(aug, name, path, groups, is_public)
    aug.save()

    with action_utils.WebserverChange() as webserver_change:
        webserver_change.enable('sharing-freedombox')


def _add(aug, name: str, path: str, groups: list[str], is_public: bool):
    """Insert a share using augeas."""
    path = '"' + path.replace('"', r'\"') + '"'
    url = '/share/' + name

    aug.set('$conf/directive[last() + 1]', 'Alias')
    aug.set('$conf/directive[last()]/arg[1]', url)
    aug.set('$conf/directive[last()]/arg[2]', path)

    aug.set('$conf/Location[last() + 1]/arg', url)

    aug.set('$conf/Location[last()]/directive[last() + 1]', 'Include')
    aug.set('$conf/Location[last()]/directive[last()]/arg',
            'includes/freedombox-sharing.conf')

    if not is_public:
        aug.set('$conf/Location[last()]/directive[last() + 1]', 'Use')
        aug.set('$conf/Location[last()]/directive[last()]/arg[1]',
                'AuthOpenIDConnect')

        for group_name in groups:
            aug.set('$conf/Location[last()]/directive[last() + 1]', 'Use')
            aug.set('$conf/Location[last()]/directive[last()]/arg[1]',
                    'RequireGroup')
            aug.set('$conf/Location[last()]/directive[last()]/arg[2]',
                    group_name)
    else:
        aug.set('$conf/Location[last()]/directive[last() + 1]', 'Require')
        aug.set('$conf/Location[last()]/directive[last()]/arg[1]', 'all')
        aug.set('$conf/Location[last()]/directive[last()]/arg[2]', 'granted')


@privileged
def remove(name: str):
    """Remove a share from Apache configuration."""
    aug = load_augeas()
    _remove(aug, name)
    aug.save()

    with action_utils.WebserverChange() as webserver_change:
        webserver_change.enable('sharing-freedombox')


def _remove(aug, name: str):
    """Remove from configuration using augeas lens."""
    url_to_remove = '/share/' + name
    for directive in aug.match('$conf/directive'):
        if aug.get(directive) != 'Alias':
            continue

        url = aug.get(directive + '/arg[1]')
        if url == url_to_remove:
            aug.remove(directive)

    for location in aug.match('$conf/Location'):
        url = aug.get(location + '/arg')
        if url == url_to_remove:
            aug.remove(location)


def _get_name_from_url(url):
    """Return the name of the share given the URL for it."""
    matches = re.match(r'/share/([a-z0-9\-]*)', url)
    if not matches:
        raise ValueError

    return matches[1]


def _list(aug=None):
    """List all Apache configuration shares."""
    if not aug:
        aug = load_augeas()

    shares = []

    for match in aug.match('$conf/directive'):
        if aug.get(match) != 'Alias':
            continue

        url = aug.get(match + '/arg[1]')
        path = aug.get(match + '/arg[2]')

        path = path.removesuffix('"').removeprefix('"')
        path = path.replace(r'\"', '"')
        try:
            name = _get_name_from_url(url)
            shares.append({
                'name': name,
                'path': path,
                'url': '/share/' + name
            })
        except ValueError:
            continue

    for location in aug.match('$conf/Location'):
        url = aug.get(location + '/arg')

        try:
            name = _get_name_from_url(url)
        except ValueError:
            continue

        groups = []
        # Old style pubtkt configuration
        for group in aug.match(location + '//directive["TKTAuthToken"]/arg'):
            groups.append(aug.get(group))

        # New style OpenID Connect configuration
        for require_group in aug.match(
                location + '//directive["Use" and arg[1] = "RequireGroup"]'):
            group = aug.get(require_group + '/arg[2]')
            groups.append(group)

        def _is_public():
            """Must contain the line 'Require all granted'."""
            require = location + '//directive["Require"]'
            return bool(aug.match(require)) and aug.get(
                require +
                '/arg[1]') == 'all' and aug.get(require +
                                                '/arg[2]') == 'granted'

        for share in shares:
            if share['name'] == name:
                share['groups'] = groups
                share['is_public'] = _is_public()

    return shares


@privileged
def list_shares() -> list[dict[str, object]]:
    """List all Apache configuration shares and print as JSON."""
    return _list()


@privileged
def uninstall():
    """Remove apache config when app is uninstalled."""
    pathlib.Path(APACHE_CONFIGURATION).unlink(missing_ok=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Views for WireGuard application.
"""

import segno
import urllib.parse

from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.http import Http404, HttpResponse
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils.translation import gettext as _
from django.views.generic import FormView, TemplateView, View

from io import BytesIO

from plinth import network
from plinth.modules.names.components import DomainName
from plinth.views import AppView

from . import forms, utils


class WireguardView(AppView):
    """Serve configuration page."""
    app_id = 'wireguard'
    diagnostics_module_name = 'wireguard'
    template_name = 'wireguard.html'

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        info = utils.get_info()
        server_info = info['my_server']
        context['server'] = server_info
        context['client_peers'] = info['my_client']['servers']
        context['server_endpoints'] = []

        if server_info:
            domains = DomainName.list_names(filter_for_service='wireguard')
            listen_port = server_info.get('listen_port')
            context['server_endpoints'] = [
                f'{domain}:{listen_port}' for domain in domains
                if not domain.endswith('.local')
            ]

        return context


class AddClientView(SuccessMessageMixin, FormView):
    """View to add a client."""
    form_class = forms.AddClientForm
    template_name = 'wireguard_add_client.html'
    success_url = reverse_lazy('wireguard:index')
    success_message = _('Added new client.')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Add Allowed Client')

        # Show next available IP.
        try:
            connection = utils._server_connection()
            setting_name = utils.nm.SETTING_WIREGUARD_SETTING_NAME
            settings = connection.get_setting_by_name(setting_name)
            context['next_ip'] = utils._get_next_available_ip_address(settings)
        except Exception:
            context['next_ip'] = None

        return context

    def form_valid(self, form):
        """Add the client."""
        public_key = form.cleaned_data.get('public_key')
        try:
            utils.add_client(public_key)
        except ValueError:
            messages.warning(self.request,
                             _('Client with public key already exists'))
            return redirect('wireguard:index')

        return super().form_valid(form)


class SessionClientDataMixin:
    """Shared session data loading for auto-client views."""

    def get_session_client_data(self, request):
        """Extract client data from session."""
        next_ip = request.session.get('next_ip')
        pubkey = request.session.get('client_pubkey')
        privkey = request.session.get('client_privkey')
        endpoint = request.session.get('endpoint')

        if not all([next_ip, privkey, pubkey, endpoint]):
            raise Http404("Session expired")

        return {
                'next_ip': next_ip,
                'privkey': privkey,
                'pubkey': pubkey,
                'endpoint': endpoint
                }

    def get_client_config(self, request):
        """Rebuild client config from session."""
        data = self.get_session_client_data(request)

        return utils.build_client_config(
                data['next_ip'], data['privkey'],
                data['pubkey'], data['endpoint']
                )


class ClientActionsView(SessionClientDataMixin, View):
    action = None

    def get(self, request):
        if self.action == 'download':
            config = self.get_client_config(request)
            response = HttpResponse(config, content_type='text/plain')
            response['Content-Disposition'] = \
                'attachment; filename="wg-client.conf"'
            return response
        elif self.action == 'qr':
            qrcode = segno.make(config)
            buffer = BytesIO()
            qrcode.save(buffer, kind='svg', scale=5)

            return HttpResponse(buffer.getvalue(),
                                content_type='image/svg+xml')

        raise Http404("Invalid action")


class AutoAddClientView(SuccessMessageMixin, FormView):
    """View to add a client with keypair generation."""
    form_class = forms.AutoAddClientForm
    template_name = 'wireguard_auto_add_client.html'
    success_url = reverse_lazy('wireguard:index')
    success_message = _('Added new client.')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Add Allowed Client')

        context['domains'] = []
        info = utils.get_info()
        server_info = info['my_server']

        if server_info:
            domains = DomainName.list_names(filter_for_service='wireguard')
            filtered_domains = [
                domain for domain in domains if not domain.endswith('.local')
            ]
            port = server_info.get('listen_port', 51820)
            endpoint = f"{filtered_domains[0]}:{port}"

        try:
            client_privkey, client_pubkey = utils.generate_client_keypair()

            # Get next IP
            connection = utils._server_connection()
            setting_name = utils.nm.SETTING_WIREGUARD_SETTING_NAME
            settings = connection.get_setting_by_name(setting_name)
            next_ip = utils._get_next_available_ip_address(settings)

            data = {
                'next_ip': next_ip,
                'client_privkey': client_privkey,
                'client_pubkey': client_pubkey,
                'endpoint': endpoint
                    }

            # Add properties to template context
            context['domains'] = filtered_domains
            context.update(data)

            # Store info on instance for reuse
            self.request.session.update(data)

        except Exception as e:
            messages.warning(f"Client key generation failed: {e}")
        pass

        return context

    def form_valid(self, form):
        """Add client using generated public key."""
        try:
            client_pubkey = self.request.session.pop('client_pubkey')
            utils.add_client(client_pubkey)
        except KeyError:
            messages.warning(self.request,
                             _('Session expired. Please try again.'))
            return redirect('wireguard:auto-add-client')
        except ValueError:
            messages.warning(self.request, _('Client already exists'))
            return redirect('wireguard:index')

        return super().form_valid(form)


class ShowClientView(SuccessMessageMixin, TemplateView):
    """View to show a client's details."""
    template_name = 'wireguard_show_client.html'

    def get_context_data(self, **kwargs) -> dict[str, object]:
        """Return additional context data for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Allowed Client')

        public_key = urllib.parse.unquote(self.kwargs['public_key'])
        server_info = utils.get_info()['my_server']
        if not server_info or public_key not in server_info['peers']:
            raise Http404

        domains = DomainName.list_names(filter_for_service='wireguard')
        context['server'] = server_info
        context['client'] = server_info['peers'][public_key]
        context['endpoints'] = [
            domain + ':' + str(server_info['listen_port'])
            for domain in domains if not domain.endswith('.local')
        ]
        return context


class EditClientView(SuccessMessageMixin, FormView):
    """View to modify a client."""
    form_class = forms.AddClientForm
    template_name = 'wireguard_edit_client.html'
    success_url = reverse_lazy('wireguard:index')
    success_message = _('Updated client.')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Modify Client')
        return context

    def get_initial(self):
        """Get initial form data."""
        initial = super().get_initial()
        initial['public_key'] = urllib.parse.unquote(self.kwargs['public_key'])
        return initial

    def form_valid(self, form):
        """Update the client."""
        old_public_key = form.initial['public_key']
        public_key = form.cleaned_data.get('public_key')

        if old_public_key != public_key:
            try:
                utils.add_client(public_key)
            except ValueError:
                messages.warning(self.request,
                                 _('Client with public key already exists'))

            utils.remove_client(old_public_key)

        return super().form_valid(form)


class DeleteClientView(SuccessMessageMixin, TemplateView):
    """View to delete a client."""
    template_name = 'wireguard_delete_client.html'

    def get_context_data(self, **kwargs):
        """Return additional context data for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Delete Allowed Client')
        context['public_key'] = urllib.parse.unquote(self.kwargs['public_key'])
        return context

    def post(self, request, public_key):
        """Delete the client."""
        public_key = urllib.parse.unquote(public_key)
        try:
            utils.remove_client(public_key)
            messages.success(request, _('Client deleted.'))
        except KeyError:
            messages.error(request, _('Client not found'))

        return redirect('wireguard:index')


class AddServerView(SuccessMessageMixin, FormView):
    """View to add a server."""
    form_class = forms.AddServerForm
    template_name = 'wireguard_add_server.html'
    success_url = reverse_lazy('wireguard:index')
    success_message = _('Added new server.')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Add Connection to Server')
        return context

    def form_valid(self, form):
        """Add the server."""
        utils.add_server(form.get_settings())
        return super().form_valid(form)


class ShowServerView(SuccessMessageMixin, TemplateView):
    """View to show a server's details."""
    template_name = 'wireguard_show_server.html'

    def get_context_data(self, **kwargs):
        """Return additional context data for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Connection to Server')

        interface = self.kwargs['interface']
        info = utils.get_info()
        server = info['my_client']['servers'].get(interface)
        if not server:
            raise Http404

        context['interface'] = interface
        context['server'] = server
        return context


class EditServerView(SuccessMessageMixin, FormView):
    """View to modify a server."""
    form_class = forms.AddServerForm
    template_name = 'wireguard_edit_server.html'
    success_url = reverse_lazy('wireguard:index')
    success_message = _('Updated server.')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Modify Connection to Server')
        return context

    def get_initial(self):
        """Get initial form data."""
        initial = super().get_initial()
        interface = self.kwargs['interface']
        info = utils.get_nm_info()
        server = info.get(interface)
        if not server:
            raise Http404

        initial['ip_address_and_network'] = server.get(
            'ip_address_and_network')
        if server['peers']:
            peer = next(peer for peer in server['peers'].values())
            initial['peer_endpoint'] = peer['endpoint']
            initial['peer_public_key'] = peer['public_key']
            initial['private_key'] = server['private_key']
            initial['preshared_key'] = peer['preshared_key']
            initial['default_route'] = server['default_route']

        return initial

    def form_valid(self, form):
        """Update the server."""
        interface = self.kwargs['interface']
        utils.edit_server(interface, form.get_settings())
        return super().form_valid(form)


class DeleteServerView(SuccessMessageMixin, TemplateView):
    """View to delete a server."""
    template_name = 'wireguard_delete_server.html'

    def get_context_data(self, **kwargs):
        """Return additional context data for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Delete Connection to Server')

        interface = self.kwargs['interface']
        info = utils.get_nm_info()
        server = info.get(interface)
        if not server:
            raise Http404

        context['interface'] = interface
        if server['peers']:
            peer = next(peer for peer in server['peers'].values())
            context['peer_endpoint'] = peer['endpoint']
            context['peer_public_key'] = peer['public_key']

        return context

    def post(self, request, interface):
        """Delete the server."""
        connection = network.get_connection_by_interface_name(interface)
        network.delete_connection(connection.get_uuid())
        messages.success(request, _('Server deleted.'))
        return redirect('wireguard:index')


class EnableServerView(SuccessMessageMixin, View):
    """View to enable the WireGuard server."""

    def post(self, request):
        """Create server interface."""
        try:
            utils.setup_server()
            messages.success(request,
                             _('WireGuard server started successfully.'))
        except Exception as error:
            messages.error(
                request,
                _('Failed to start WireGuard server: {}').format(error))
        return redirect('wireguard:index')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for upgrades.
"""

from django.utils.translation import gettext_lazy as _

backup = {'config': {'files': ['/etc/apt/apt.conf.d/20auto-upgrades']}}

tags = [_('Security'), _('Automatic'), _('Reboots'), _('New features')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for System Monitoring (cockpit-pcp) in ‘System’.
"""

from django.urls import re_path

from plinth.views import AppView

urlpatterns = [
    re_path(r'^sys/performance/$', AppView.as_view(app_id='performance'),
            name='index'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for pagekite.
"""

from django.utils.translation import gettext_lazy as _

backup = {
    'config': {
        'directories': ['/etc/pagekite.d/']
    },
    'services': ['pagekite']
}

tags = [
    _('Reachability'),
    _('Remote access'),
    _('Tunneling'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for bepasty app.
"""

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.bepasty]


class TestBepastyApp(functional.BaseAppTests):
    app_name = 'bepasty'
    has_service = False
    has_web = True

    def test_set_default_permissions_list_and_read_all(self, session_browser):
        functional.app_enable(session_browser, 'bepasty')
        _logout(session_browser)
        _set_default_permissions(session_browser, 'read list')

        assert _can_list_all(session_browser)

    def test_set_default_permissions_read_files(self, session_browser):
        functional.app_enable(session_browser, 'bepasty')
        _logout(session_browser)
        _set_default_permissions(session_browser, 'read')

        assert _cannot_list_all(session_browser)

    def test_add_password(self, session_browser):
        functional.app_enable(session_browser, 'bepasty')
        password_added = _add_and_save_password(session_browser)

        assert _can_login(session_browser, password_added)

    def test_remove_password(self, session_browser):
        functional.app_enable(session_browser, 'bepasty')
        password_added = _add_and_save_password(session_browser)
        _remove_all_passwords(session_browser)

        assert not _can_login(session_browser, password_added)

    @pytest.mark.backups
    def test_backup_restore(self, session_browser):
        functional.app_enable(session_browser, 'bepasty')
        password_added = _add_and_save_password(session_browser)
        functional.backup_create(session_browser, 'bepasty', 'test_bepasty')

        _remove_all_passwords(session_browser)
        functional.backup_restore(session_browser, 'bepasty', 'test_bepasty')

        assert functional.is_available(session_browser, 'bepasty')
        assert _can_login(session_browser, password_added)


def _add_and_save_password(session_browser):
    _remove_all_passwords(session_browser)
    _add_password(session_browser)
    return _get_password(session_browser)


def _set_default_permissions(browser, permissions=''):
    functional.nav_to_module(browser, 'bepasty')
    browser.choose('default_permissions', permissions)
    functional.submit(browser, form_class='form-configuration')


def _add_password(browser):
    functional.visit(browser, '/freedombox/apps/bepasty/add/')
    for permission in ['read', 'create', 'list', 'delete', 'admin']:
        browser.find_by_css('#id_bepasty-permissions input[value="{}"]'.format(
            permission)).check()
    browser.fill('bepasty-comment', 'bepasty functional test')
    functional.submit(browser, form_class='form-bepasty')


def _remove_all_passwords(browser):
    functional.nav_to_module(browser, 'bepasty')
    while True:
        remove_button = browser.find_by_css('.password-remove')
        if remove_button:
            functional.submit(browser, element=remove_button)
        else:
            break


def _get_password(browser):
    functional.nav_to_module(browser, 'bepasty')
    return browser.find_by_css('.password-password').first.text


def _can_login(browser, password):
    _logout(browser)
    browser.fill('token', password)
    login = browser.find_by_xpath('//form//button')
    functional.submit(browser, element=login)

    return bool(browser.find_by_value('Logout'))


def _logout(browser):
    functional.visit(browser, '/bepasty')
    logout = browser.find_by_value('Logout')
    if logout:
        logout.click()


def _can_list_all(browser):
    functional.visit(browser, '/bepasty')
    return functional.eventually(browser.links.find_by_href,
                                 ['/bepasty/+list'], 5)


def _cannot_list_all(browser):
    functional.visit(browser, '/bepasty/+list')
    return functional.eventually(browser.is_text_present, ['Forbidden'], 5)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configuration helper for Gitweb."""

import configparser
import logging
import os
import pathlib
import re
import shutil
import subprocess
import time
from typing import Any
from urllib import parse

from plinth import action_utils
from plinth.actions import privileged, secret_str
from plinth.modules.gitweb.forms import RepositoryValidator, get_name_from_url
from plinth.modules.gitweb.manifest import GIT_REPO_PATH, REPO_DIR_OWNER

logger = logging.getLogger(__name__)


def validate_repo_name(name: str) -> str:
    """Validate a repository name and add .git extension if necessary."""
    RepositoryValidator()(name)
    if not name.endswith('.git'):
        name = name + '.git'

    return name


def validate_repo_url(url: secret_str) -> secret_str:
    """Validate a repository URL."""
    RepositoryValidator(input_should_be='url')(url)
    return url


@privileged
def setup():
    """Configure Gitweb module."""
    # Disable default Apache2 Gitweb configuration.
    action_utils.webserver_disable('gitweb')

    # Configure Git client.
    if not _get_git_global_config('init.defaultBranch'):
        _set_git_global_config('init.defaultBranch', 'main')
    if not _get_git_global_config('credential.helper'):
        _set_git_global_config('credential.helper', 'cache')


def _get_git_global_config(key: str) -> str | None:
    """Return a value from Git global configuration."""
    try:
        value = action_utils.run(['git', 'config', '--global', '--get', key],
                                 check=True).stdout.decode().strip()
    except subprocess.CalledProcessError as exception:
        if exception.returncode == 1:  # Configuration option doesn't exist
            return None
        raise

    return value


def _set_git_global_config(key: str, value: str) -> None:
    """Set a Git global configuration value."""
    action_utils.run(['git', 'config', '--global', key, value], check=True)


def _setup_git_credentials(url: secret_str) -> str:
    """Set up git credential helper and return URL without credentials."""
    url_parts = parse.urlsplit(url)
    safe_netloc = url_parts.netloc.split('@')[-1]
    safe_url = url_parts._replace(netloc=safe_netloc).geturl()
    username = url_parts.username or ''
    password = url_parts.password or ''

    if username or password:
        # Feed credentials to Git credential helper
        input = (f'protocol={url_parts.scheme}\n'
                 f'host={safe_netloc}\n'
                 f'username={username}\n'
                 f'password={password}\n\n')
        env = dict(os.environ, GIT_TERMINAL_PROMPT='0')
        action_utils.run(['git', 'credential', 'approve'],
                         input=input.encode(), stdout=subprocess.DEVNULL,
                         check=True, env=env)

    return safe_url


def _clone_with_progress_report(url: secret_str, repo_dir: pathlib.Path):
    """Clone a repository and write progress info to the file."""
    starttime = time.time()
    status_file = repo_dir / 'clone_progress'
    repo_temp_dir = repo_dir / '.temp'
    # do not ask for credidentials and set low speed timeout
    env = dict(os.environ, GIT_TERMINAL_PROMPT='0', LC_ALL='C',
               GIT_HTTP_LOW_SPEED_LIMIT='100', GIT_HTTP_LOW_SPEED_TIME='60')

    safe_url = _setup_git_credentials(url)
    logger.info(f'Cloning Git repository {safe_url} ...')
    proc = subprocess.Popen(
        ['git', 'clone', '--bare', '--progress', safe_url,
         str(repo_temp_dir)], stderr=subprocess.PIPE, text=True, env=env)
    assert proc.stderr is not None

    # write clone progress to the file
    errors = []
    while True:
        line = proc.stderr.readline()
        if not line:
            break

        if 'error:' in line or 'fatal:' in line:
            errors.append(line.strip())

        currenttime = time.time()
        if currenttime - starttime > 1:
            elapsed = _clone_status_line_to_percent(line)
            if elapsed is not None:
                try:
                    status_file.write_text(elapsed)
                except OSError as error:
                    errors.append(str(error))

            starttime = currenttime

    # make sure process is ended
    try:
        proc.communicate(timeout=10)
    except subprocess.TimeoutExpired:
        proc.kill()

    status_file.unlink()
    if proc.returncode != 0:
        shutil.rmtree(repo_dir)
        logger.error('Git repository cloning failed: %s', '\n'.join(errors))
        raise RuntimeError('Git repository cloning failed.', errors)


def _prepare_clone_repo(url: secret_str, is_private: bool):
    """Prepare cloning a repository."""
    repo_name = get_name_from_url(url)
    if not repo_name.endswith('.git'):
        repo_name = repo_name + '.git'

    repo_dir = GIT_REPO_PATH / repo_name
    os.mkdir(repo_dir)

    status_file = repo_dir / 'clone_progress'
    try:
        if is_private:
            _set_access_status(repo_name, 'private')

        status_file.write_text('0')
    except OSError:
        shutil.rmtree(repo_dir)
        raise


def _clone_status_line_to_percent(line):
    """Parse Git clone command output."""
    result = re.match(r'.* ([0-9]+)% ', line)
    if result is not None:
        text = result.group(0)
        progress = int(result.group(1))
        if 'Counting objects' in text:
            total_progress = 0.05 * progress
        elif 'Compressing objects' in text:
            total_progress = 5 + 0.05 * progress
        elif 'Receiving objects' in text:
            total_progress = 10 + 0.6 * progress
        elif 'Resolving deltas' in text:
            total_progress = 70 + 0.3 * progress
        else:
            return None

        return str(int(total_progress))

    return None


def _clone_repo(url: secret_str, description: str, owner: str,
                keep_ownership: bool):
    """Clone a repository."""
    repo = get_name_from_url(url)
    if not repo.endswith('.git'):
        repo = repo + '.git'

    repo_path = GIT_REPO_PATH / repo
    repo_temp_path = repo_path / '.temp'

    _clone_with_progress_report(url, repo_path)

    for item in os.listdir(repo_temp_path):
        shutil.move(os.path.join(repo_temp_path, item), repo_path)

    shutil.rmtree(repo_temp_path)
    if not keep_ownership:
        action_utils.run(
            ['chown', '-R', f'{REPO_DIR_OWNER}:{REPO_DIR_OWNER}', repo],
            cwd=GIT_REPO_PATH, check=True)

    _set_repo_description(repo, description)
    _set_repo_owner(repo, owner)


def _create_repo(repo: str, description: str, owner: str, is_private: bool,
                 keep_ownership: bool):
    """Create an empty repository."""
    try:
        action_utils.run(['git', 'init', '-q', '--bare', repo],
                         cwd=GIT_REPO_PATH, check=True)
        if not keep_ownership:
            action_utils.run(
                ['chown', '-R', f'{REPO_DIR_OWNER}:{REPO_DIR_OWNER}', repo],
                cwd=GIT_REPO_PATH, check=True)
        _set_repo_description(repo, description)
        _set_repo_owner(repo, owner)
        if is_private:
            _set_access_status(repo, 'private')
    except (subprocess.CalledProcessError, OSError):
        repo_path = GIT_REPO_PATH / repo
        if repo_path.is_dir():
            shutil.rmtree(repo_path)
        raise


def _get_default_branch(repo):
    """Get default branch of the repository."""
    repo_path = GIT_REPO_PATH / repo

    return action_utils.run_as_user(
        ['git', '-C',
         str(repo_path), 'symbolic-ref', '--short', 'HEAD'],
        username=REPO_DIR_OWNER, check=True).stdout.decode().strip()


def _get_repo_description(repo):
    """Set description of the repository."""
    description_file = GIT_REPO_PATH / repo / 'description'
    if description_file.exists():
        description = description_file.read_text()
    else:
        description = ''

    return description


def _set_repo_description(repo, description):
    """Set description of the repository."""
    description_file = GIT_REPO_PATH / repo / 'description'
    description_file.write_text(description)


def _get_repo_owner(repo):
    """Set repository's owner name."""
    repo_config = GIT_REPO_PATH / repo / 'config'
    config = configparser.ConfigParser()
    config.read(repo_config)
    try:
        owner = config['gitweb']['owner']
    except KeyError:
        owner = ''

    return owner


def _set_repo_owner(repo, owner):
    """Set repository's owner name."""
    repo_config = GIT_REPO_PATH / repo / 'config'
    config = configparser.ConfigParser()
    config.read(repo_config)
    if not config.has_section('gitweb'):
        config.add_section('gitweb')

    config['gitweb']['owner'] = owner
    with repo_config.open('w', encoding='utf-8') as file_handle:
        config.write(file_handle)


def _get_access_status(repo):
    """Get repository's access status."""
    private_file = GIT_REPO_PATH / repo / 'private'
    if private_file.exists():
        return 'private'

    return 'public'


def _set_access_status(repo, status):
    """Set repository as private or public."""
    private_file = GIT_REPO_PATH / repo / 'private'
    if status == 'private':
        private_file.touch()
    elif status == 'public':
        private_file.unlink(missing_ok=True)


def _get_branches(repo):
    """Return list of the branches in the repository."""
    process = action_utils.run_as_user(
        ['git', '-C', repo, 'branch', '--format=%(refname:short)'],
        cwd=GIT_REPO_PATH, username=REPO_DIR_OWNER, check=True)

    return process.stdout.decode().strip().split()


@privileged
def get_branches(name: str) -> dict[str, Any]:
    """Check whether a branch exists in the repository."""
    repo = validate_repo_name(name)
    return dict(default_branch=_get_default_branch(repo),
                branches=_get_branches(repo))


@privileged
def rename_repo(old_name: str, new_name: str):
    """Rename a repository."""
    old_name = validate_repo_name(old_name)
    new_name = validate_repo_name(new_name)
    oldpath = GIT_REPO_PATH / old_name
    newpath = GIT_REPO_PATH / new_name
    oldpath.rename(newpath)


@privileged
def set_default_branch(name: str, branch: str):
    """Set description of the repository."""
    repo = validate_repo_name(name)
    if branch not in _get_branches(repo):
        raise ValueError('No such branch')

    action_utils.run_as_user(
        ['git', '-C', repo, 'symbolic-ref', 'HEAD', f'refs/heads/{branch}'],
        cwd=GIT_REPO_PATH, check=True, username=REPO_DIR_OWNER)


@privileged
def set_repo_description(name: str, description: str):
    """Set description of the repository."""
    repo = validate_repo_name(name)
    _set_repo_description(repo, description)


@privileged
def set_repo_owner(name: str, owner: str):
    """Set repository's owner name."""
    repo = validate_repo_name(name)
    _set_repo_owner(repo, owner)


@privileged
def set_repo_access(name: str, access: str):
    """Set repository's access status."""
    repo = validate_repo_name(name)
    if access not in ('public', 'private'):
        raise ValueError('Invalid access parameter')

    _set_access_status(repo, access)


@privileged
def repo_info(name: str) -> dict[str, str]:
    """Get information about repository."""
    repo = validate_repo_name(name)
    repo_path = GIT_REPO_PATH / repo
    if not repo_path.exists():
        raise RuntimeError('Repository not found')

    return dict(name=repo[:-4], description=_get_repo_description(repo),
                owner=_get_repo_owner(repo), access=_get_access_status(repo),
                default_branch=_get_default_branch(repo))


@privileged
def create_repo(url: secret_str | None = None, name: str | None = None,
                description: str = '', owner: str = '',
                keep_ownership: bool = False, is_private: bool = False,
                skip_prepare: bool = False, prepare_only: bool = False):
    """Create a new or clone a remote repository."""
    if url:
        url = validate_repo_url(url)

    if name:
        repo = validate_repo_name(name)

    if url:
        if not skip_prepare:
            _prepare_clone_repo(url, is_private)

        if not prepare_only:
            _clone_repo(url, description, owner, keep_ownership)
    elif repo is not None:
        _create_repo(repo, description, owner, is_private, keep_ownership)


@privileged
def repo_exists(url: secret_str) -> bool:
    """Return whether remote repository exists."""
    url = validate_repo_url(url)
    safe_url = _setup_git_credentials(url)
    env = dict(os.environ, GIT_TERMINAL_PROMPT='0')
    try:
        action_utils.run(['git', 'ls-remote', safe_url, 'HEAD'], timeout=10,
                         env=env, check=True)
        return True
    except subprocess.CalledProcessError:
        return False


@privileged
def delete_repo(name: str):
    """Delete a git repository."""
    repo = validate_repo_name(name)
    repo_path = GIT_REPO_PATH / repo
    shutil.rmtree(repo_path)


@privileged
def uninstall():
    """Remove git repositories."""
    for item in GIT_REPO_PATH.iterdir():
        shutil.rmtree(item, ignore_errors=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""URLs for the Privacy module."""

from django.urls import re_path

from .views import PrivacyAppView

urlpatterns = [
    re_path(r'^sys/privacy/$', PrivacyAppView.as_view(), name='index'),
]



"""FreedomBox app for implementing a OpenID Connect Provider.

With this app, FreedomBox implements a full OpenID Connect Provider along with
OpenID Discovery. Only authorization code grant type is currently supported but
can be easily extended to support other grant types if necessary. See this code
comment for quick understand of the flow works:
https://github.com/oauthlib/oauthlib/blob/master/oauthlib/oauth2/rfc6749/grant_types/authorization_code.py#L64

In the list of OpenID Connect claims provided to Relying Party, we override
django-oauth-toolkit's default 'sub' claim from being the user ID to username.
Additionally, we also provide basic profile information and 'freedombox_groups'
with list of all group names that the user account is part of.

"Clients" or "Applications": must be registered before they can be used with
the Identity Provider. Dynamic Client Registration is not supported yet. The
OpenIDConnect FreedomBox component will help with this registration.

Redirect URLs: After authorization is provided by the user, the URL of the
application to redirect to must be verified by the Identity Provider. This is
usually provided at the time of client registration. However, in FreedomBox,
since list of allowed domains keeps changing, the list of allowed redirect URLs
must keep changing as well. The OpenIDConnect component will also help with
that. Finally, there is overridden verification logic that ensures that
accessing protected applications using IP addresses or localhost domain names
in URLs is allowed.

The implement is done by implementing all the URLs in /freedombox/o. Most of
the implementation for these views and models is provided by
django-oauth-toolkit.
"""

import logging

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module

from . import components

logger = logging.getLogger(__name__)


class OIDCApp(app_module.App):
    """FreedomBox app for OpenID Connect Provider."""

    app_id = 'oidc'

    _version = 1

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               is_essential=True,
                               name=_('OpenID Connect Provider'))
        self.add(info)


def on_domain_added(sender: str, domain_type: str, name: str = '',
                    description: str = '',
                    services: str | list[str] | None = None, **kwargs):
    """Add domain to global list."""
    if not domain_type or not name:
        return

    logger.info('Updating all OpenID Connect components for domain add - %s',
                name)
    components.OpenIDConnect.update_domains_for_all()


def on_domain_removed(sender: str, domain_type: str, name: str = '', **kwargs):
    """Remove domain from global list."""
    if not domain_type or not name:
        return

    logger.info(
        'Updating all OpenID Connect components for domain remove - %s', name)
    components.OpenIDConnect.update_domains_for_all()
